/*
  File :     cupsprintmi.c


  Author :   Francesco Prelz ($Author: printmi $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
  09-Sep-2010 Created.

  Description: Print client wrapper that is suitable
               for using as a back-end for CUPS (the "common" Unix
               printing system). 
 */

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>

#include "print.h"

#define CUPS_BACKEND_OK     0
#define CUPS_BACKEND_FAILED 1
#define CUPS_BACKEND_HOLD   3

#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX 256
#endif

#define WILDCARD_UID 999999

#define DEFAULT_DOMAIN "mi.infn.it"

char *
get_secondary_from_primary(char *prserv)
{
  char *dloc;

  if ((dloc = strchr(prserv,'.')) == NULL)
   {
    return strdup(SERV_HOST_NAME_II);
   }
}

int
list_queues_for_cups()
 {
  int sockfd;
  struct control param;
  char recvline[MAXLINE + 1];
  char hostname[HOST_NAME_MAX];
  char *fqsrv=NULL;
  char *fqsec;
  char *piploc;
  char *dotl;
  int recvl;
  int n;

  /* When run with no arguments, the backend should list the devices and 
   * schemes it supports or is advertising to stdout. 
   * The output consists of zero or more lines consisting of any 
   * of the following forms:
   * device-class scheme "Unknown" "device-info"
   * device-class device-uri "device-make-and-model" "device-info"
   * (...)
   */

  gethostname(hostname, sizeof(hostname));

  /* Get connection to server */
  if (does_resolve(SERV_HOST_NAME_I)) fqsrv = strdup(SERV_HOST_NAME_I);
  else
   {
    if ((dotl = strchr(hostname, '.')) != NULL)
     {
      fqsrv = (char *)malloc(strlen(SERV_HOST_NAME_I)+strlen(dotl)+2);
      if (fqsrv == NULL) return -1;
      sprintf(fqsrv,"%s%s",SERV_HOST_NAME_I,dotl);
     }
    if (fqsrv == NULL || (! does_resolve(fqsrv)))
     {
      if (fqsrv != NULL) free(fqsrv);
      fqsrv = (char *)malloc(strlen(SERV_HOST_NAME_I)+strlen(DEFAULT_DOMAIN)+2);
      if (fqsrv == NULL) return -1;
      sprintf(fqsrv,"%s.%s",SERV_HOST_NAME_I,DEFAULT_DOMAIN);
      if (! does_resolve(fqsrv))
       {
        free(fqsrv);
        return -1;
       }
     }
   }

  if (fqsrv == NULL) return -1;

  if ((is_alive(fqsrv)) == 0) /*Primary serv. down*/
   {
    fqsec=get_secondary_from_primary(fqsrv);
    if (fqsec == NULL)
     {
      free(fqsrv);
      return -1;
     }
    if ((is_alive(fqsec)) == 0)  /*Secondary serv. down*/
     {
      free(fqsrv);
      free(fqsec);
      return -1;
     }
    sockfd=sock_init(fqsec);
    free(fqsec);
   }
  else sockfd=sock_init(fqsrv);

  /* Prepare lpq request */
  snprintf(param.user, sizeof(param.user), "Cups user on %s.", hostname);
  param.host[sizeof(param.host)-1] = '\000';
  strncpy(param.host, hostname, sizeof(param.host));
  *param.file='\000';
  *param.entry='\000';
  param.uid = WILDCARD_UID;
  param.reserve=0;
  param.id_only=1;
  param.copies=0;

  str_ctr(sockfd, "lpq", &param, FALSE);
  while( (n = readline(sockfd, recvline, MAXLINE)) != 0) 
   {
    if (n < 0 ) return -1;
    recvl = strlen(recvline) - 1;
    if (recvline[recvl] == '\n')
     {
      recvline[recvl] = '\000';
     }
    piploc = strchr(recvline, '|');
    if (piploc != NULL)
     {
      *piploc='\000';
      piploc++; 
      printf("network cupsprintmi://%s:%d/%s \"Generic %s(%s)\" \"%s\" " 
             "\"CMD:POSTSCRIPT,PDF\" \"%s\"\n",
             fqsrv,
             SERV_TCP_PORT, recvline, 
             recvline, piploc, recvline,
             piploc);
     }
   }

  free(fqsrv);
  close(sockfd);
  return 0;
 }

char *
grab_cups_option(char *opname, char *opstring)
 {
   char *opf;
   char *nsp;
   char *ret;
   int atlen;
 
   opf = strstr(opstring,opname);
 
   if (opf == NULL) return NULL;
 
   opf+=strlen(opname);
   if (*opf != '=') return NULL;
   opf++;
 
   nsp = strchr(opf, ' ');
   if (nsp == NULL) return strdup(opf);
    
   atlen = (int)(nsp-opf);
   ret = (char *)malloc(atlen+1); 
   if (ret == NULL) return NULL;
 
   strncpy(ret,opf,atlen);
   ret[atlen]='\000';
 
   return ret; 
 }

int
append_printmi_option(char **options, char *new_opt)
 {
  int old_len, new_len;
  char *new_opts;

  if (*options == NULL) old_len=0;
  else old_len = strlen(*options);

  new_len = old_len+strlen(new_opt)+1; 

  new_opts = (char *)realloc(*options, new_len);
  if (new_opts == NULL) return -1;

  if (old_len == 0) strcpy(new_opts, new_opt);
  else
   {
    strcat(new_opts," ");
    strcat(new_opts,new_opt);
   }

  *options = new_opts;
  return 0;
 }

int main (int argc, char *argv[])
 {
  int i,j,size,count;
  FILE *input;
  char *in_file;
  char *cups_id;
  char *user;
  char *uri, *queue;
  char *copt;
  char *cval;
  char *popt;
  char *uri_env;
  char *fqsrv, *fqsec;
  char *host_start, *host_end;
  int host_len;
  int n_copies;
  int need_to_close_input = FALSE;
  char hostname[HOST_NAME_MAX];

  struct control param;

  int sockfd=-1;

  if (argc < 2)
   {
    if (list_queues_for_cups() < 0) exit(CUPS_BACKEND_FAILED);
    else                            exit(CUPS_BACKEND_OK);
   }

  if (argc < 6)
   {
    fprintf(stderr,"Usage: cupsprintmi (argv[0] to be set to the printer URI) <job> <user> <title> <num-copies> <options> [filename]\n");
    exit(CUPS_BACKEND_FAILED);
   }

  uri = argv[0];

  if (strstr(uri,"://") != NULL)
   {
    if (strncmp(uri,"cupsprintmi://",13) != 0)
     {
      fprintf(stderr,"cupsprintmi: argv[0] should be set to an URI starting with cupsprintmi:\n");
      fprintf(stderr,"cupsprintmi: argv[0]==\"%s\"\n", argv[0]);
      exit(CUPS_BACKEND_FAILED);
     }
   }

  queue = strrchr(uri,'/');
  if (queue != NULL) queue++;
  else queue=uri;

  if (strlen(queue) <= 0)
   {
    fprintf(stderr,"cupsprintmi: invalid queue \"%s\".\n",queue);
    exit(CUPS_BACKEND_FAILED);
   }

  cups_id = argv[1];
  user = argv[2];
  n_copies = atoi(argv[4]);
  copt = argv[5];

  if (argc > 6)
   {
    in_file = argv[6];
    input = fopen(in_file, "r");
    if (input == NULL)
     {
      fprintf(stderr,"cupsprintmi: Error reading from input file %s: %s\n",
              in_file, strerror(errno));
      exit(CUPS_BACKEND_FAILED);
     }
    need_to_close_input = TRUE;
   }
  else input = stdin;

  if ((uri_env = getenv("DEVICE_URI")) != NULL)
   {
    host_start = strstr(uri_env, "://");
    if (host_start != NULL)
     {
      host_start += 3;
      host_end = strpbrk(host_start,":/");
      if (host_end == NULL) fqsrv = strdup(host_start);
      else
       {
        host_len = (int)(host_end - host_start);
        fqsrv = (char *)malloc(host_len +1);
        if (fqsrv != NULL)
         {
          strncpy(fqsrv,host_start,host_len);
          fqsrv[host_len]='\000';
          if ((is_alive(fqsrv)) == 0) /* FQ pri serv. down */
           {
            fqsec=get_secondary_from_primary(fqsrv);
            if (fqsec != NULL)
             {
              if ((is_alive(fqsec)) != 0) /* FQ sec serv. up */
               {
                sockfd=sock_init(fqsec);
               }
              free(fqsec);
             }
           }
          sockfd=sock_init(fqsrv);
          free(fqsrv);
         }
       }
     }    
   }

  if (sockfd < 0)
   {
    /* Cannot find hostname in the URI. Use default names. */
    if ((is_alive(SERV_HOST_NAME_I)) == 0) /*Primary serv. down*/
     {
      if ((is_alive(SERV_HOST_NAME_II)) == 0)  /*Secondary serv. down*/
       {
        fprintf(stderr,"cupsprintmi: Cannot connect to printmi servers: %s\n",strerror(errno));
        exit(CUPS_BACKEND_HOLD);
       }
      sockfd=sock_init(SERV_HOST_NAME_II);
     } else {
      sockfd=sock_init(SERV_HOST_NAME_I);
     }
   }

  /* The -@ flag is not used by the regular client: it will instruct
     the print server not to attempt any formatting */
  popt = strdup("-@");

  if (popt == NULL)
   {
    fprintf(stderr,"cupsprintmi: Out of memory.\n",queue);
    exit(CUPS_BACKEND_HOLD);
   }

  /* Convert a few CUPS options we know how to handle. */
  if ((cval = grab_cups_option("sides",copt)) != NULL)
   {
     if      (strcmp(cval,"two-sided-long-edge")==0) 
               append_printmi_option(&popt, "-K2");
     else if (strcmp(cval,"two-sided-short-edge")==0) 
               append_printmi_option(&popt, "-K2 -ton");

   }

  param.file[sizeof(param.file)-1] = '\000';
  strncpy(param.file,cups_id,sizeof(param.file));
  param.user[sizeof(param.user)-1] = '\000';
  strncpy(param.user,user,sizeof(param.user));
  param.host[sizeof(param.host)-1] = '\000';
  gethostname(hostname, sizeof(hostname));
  strncpy(param.host,hostname,sizeof(param.host));
  param.printer[sizeof(param.printer)-1] = '\000';
  strncpy(param.printer,queue,sizeof(param.printer));
  param.uid = WILDCARD_UID;
  param.reserve=0;
  param.id_only=0;
  param.copies=n_copies;
  *param.entry='\000';

  str_print(input, sockfd, "lpr", popt, &param, TRUE);

  close(sockfd);
  if (need_to_close_input) fclose(input);
  free(popt);

  exit(CUPS_BACKEND_OK);
 } 
