/*
  File :     lpd_client.c


  Author :   Francesco Prelz ($Author: printmi $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
   6-Oct-1997 Original release
  10-Feb-1998 Added shorter connection timeout (LPD_CONNECT_TIMEOUT)
              (via select()).
  10-Sep-1998 Added preprocessor switches not to print banner page (can
              be activated by defining PRINTMI_GENERATED_BANNER, and
              will cause the banner page to be added by the print server).
  14-Sep-1998 Added watchdog timer to terminate stuck connection
              (after a weird incident with a Lexmark Optra 1250 printer
               which would seize a connection after reporting a "no entries"
               status).
   4-Aug-2009 Added support for IPv6 via getaddrinfo().

  Description: Simple lpd client library with the ability of
               spooling a file to a remote server from a forked
               subprocess.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>

#include "lpd_client.h"

#define LPD_REMOVE_JOB(queue,user,number) \
  query = (char *)realloc(query,strlen(queue)+strlen(user)+16); \
  sprintf(query,"%c%s %s %d\n",5,queue,user,number);
#define LPD_SEND_QUEUE_STATE_L(queue) \
  query = (char *)realloc(query,strlen(queue)+4); \
  sprintf(query,"%c%s \n",4,queue);
#define LPD_SEND_QUEUE_STATE_S(queue) \
  query = (char *)realloc(query,strlen(queue)+4); \
  sprintf(query,"%c%s \n",3,queue);
#define LPD_RECEIVE_PRINTER_JOB(queue) \
  query = (char *)realloc(query,strlen(queue)+3); \
  sprintf(query,"%c%s\n",2,queue);
#define LPD_RECEIVE_CONTROL_FILE(size,name) \
  query = (char *)realloc(query,strlen(name)+25); \
  sprintf(query,"%c%d %s\n",2,size,name);
#define LPD_RECEIVE_DATA_FILE(size,name) \
  query = (char *)realloc(query,strlen(name)+25); \
  sprintf(query,"%c%d %s\n",3,size,name);
  
#define LPD_SHUTDOWN \
     shutdown(lpd_socket,SHUT_RDWR); \
     close(lpd_socket); \
     free(query);
#define STOP_TIMER \
     signal(SIGALRM, SIG_DFL); \
     setitimer(ITIMER_REAL, &watchdog_timer, NULL);

extern int _prints_debug;

jmp_buf watchdog_env;
int     watchdog_jump_initialized = FALSE;

/* Test program */

#ifdef TEST_LIBRARY
#include <pwd.h>

int _prints_debug = 1;

int main(int argc, char *argv[])
{
  char buffer[1024];
  int fd;
  int retcod;
  char host_name[80];
  struct passwd *user_data;
  char header[2];
  int file_type;
  char *lpd_server;

#define PRINTER_TO_ACCESS argv[2]

  if (argc<3)
   {
    fprintf(stderr,"Usage: %s <server> <printer>\n",argv[0]);
    exit(1);
   }
   
  lpd_server = argv[1];

  retcod = lpd_client_get_queue_state(lpd_server,PRINTER_TO_ACCESS,buffer,sizeof(buffer));
  if (retcod > 0)
    printf("Found printer %s on server %s:\n %s\n",PRINTER_TO_ACCESS,lpd_server,buffer);
  else
   {
    fprintf(stderr,"Printer %s on server %s: %s is not responding.\n",
            PRINTER_TO_ACCESS,lpd_server);
    exit(1);
   }

  if ((fd = open(argv[3],O_RDONLY))==-1)
   {
    fprintf(stderr,"Cannot open file %s.\n",argv[3]);
    exit(1);
   }
  read(fd,header,sizeof(header));

  if (memcmp(header,"%!",2)==0) file_type = LPD_POSTSCRIPT_TYPE;
  else                          file_type = 0;
  
  gethostname(host_name,sizeof(host_name));
  user_data = getpwuid(getuid());
     
  lpd_client_spool_file(lpd_server,PRINTER_TO_ACCESS,host_name,user_data->pw_name,argv[3], argv[3], file_type, 1, 0);

  retcod = lpd_client_get_queue_state(lpd_server,PRINTER_TO_ACCESS,buffer,sizeof(buffer));
  if (retcod > 0)
    printf("Found printer %s on server %s:\n %s\n",PRINTER_TO_ACCESS,lpd_server,buffer);

  exit(0);

 }
#endif

int lpd_client_spool_file(char *lpd_server, char *queue,
    char *host, char *user, char *orig_name, char *name, 
    int file_type, int queue_no, int banner)
 {
  int file_to_spool;
  struct stat file_stat;
  int lpd_socket;
  int length_to_send,banner_size;
  char *query=NULL;
  char *cont_file;
  char cont_file_name[FILENAME_MAX];
  char data_file_name[FILENAME_MAX];
  unsigned short query_length;
  int read_length,write_length,total_length=0;
  unsigned char buffer[8192];
  int retcod;
  char zero='\000';

  struct itimerval watchdog_timer;

  watchdog_timer.it_interval.tv_sec  = 0;
  watchdog_timer.it_interval.tv_usec = 0;
  watchdog_timer.it_value.tv_sec  = LPD_CLIENT_WATCHDOG_TIMEOUT;
  watchdog_timer.it_value.tv_usec = 0;
  watchdog_jump_initialized = FALSE;

  if ((file_to_spool = open(name,O_RDONLY))==-1)
   {
    fprintf(stderr,"Cannot open file %s:%s\n",name,strerror(errno));
    return(-1);
   }
   
  if (fstat(file_to_spool,&file_stat)==-1)
   {
    fprintf(stderr,"Cannot get info about file %s:%s\n",name,strerror(errno));
    return(-1);
   }
   
  /*Start watchdog timer - we don't require it to succeed*/
  signal (SIGALRM, lpd_client_hard_timeout);
  setitimer(ITIMER_REAL, &watchdog_timer, NULL);
  watchdog_timer.it_value.tv_sec  = 0;

  while ((lpd_socket = get_lpd_connection(lpd_server))<0)
   {
    if (errno!=EADDRINUSE) return(lpd_socket);
    sleep(LPD_WAIT_FOR_SOCKET);
   }

  /*Handle watchdog timeout*/
  if (setjmp(watchdog_env) > 0)
   {
      LPD_SHUTDOWN;
      signal(SIGALRM, SIG_DFL);
      return(-1);
   }
  watchdog_jump_initialized = TRUE;

  LPD_RECEIVE_PRINTER_JOB(queue);
  if (write(lpd_socket,query,strlen(query))>0)
   {
    retcod=read_lpd_ack(lpd_socket);
    if (_prints_debug != 0)
      printf("Ack to \"receive\" command: %d\n",retcod);  
    if (retcod!=0)
     {
      LPD_SHUTDOWN;
      STOP_TIMER;
      return(-1);
     }
   }
  else
   {
    fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    STOP_TIMER;
    return(-1);
   }

  sprintf(data_file_name,"dfA%03d%s",queue_no,host);
  sprintf(cont_file_name,"cfA%03d%s",queue_no,host);

  length_to_send = file_stat.st_size;

  LPD_RECEIVE_DATA_FILE(length_to_send,data_file_name);
  if (_prints_debug != 0)
    printf("Receive data query: <%s>\n",query);
  if (write(lpd_socket,query,strlen(query))>0)
   {
    retcod=read_lpd_ack(lpd_socket);
    if (_prints_debug != 0)
      printf("Ack to \"receive dat\" command: %d\n",retcod);  
    if (retcod!=0)
     {
      LPD_SHUTDOWN;
      STOP_TIMER;
      return(-1);
     }

   }
  else
   {
    fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    STOP_TIMER;
    return(-1);
   }

  while((read_length=read(file_to_spool,buffer,sizeof(buffer)))>0)
   {
    write_length = write(lpd_socket,buffer,read_length);
    if (write_length < 0)
     {
      fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
              strerror(errno));
      LPD_SHUTDOWN;
      STOP_TIMER;
      return(-1);
     }
    total_length+=write_length;
   }

  close(file_to_spool);
   
  if (total_length!=length_to_send) fprintf(stderr,"File %s: size %d sent %d\n",
                                            name,length_to_send,total_length);
  write(lpd_socket,&zero,1);
  retcod=read_lpd_ack(lpd_socket);
  if (_prints_debug != 0)
    printf("Ack to data file delivery: %d\n",retcod);  
  if (retcod!=0)
   {
    LPD_SHUTDOWN;
    STOP_TIMER;
    return(-1);
   }

  cont_file = compose_lpd_control_file(host, user, orig_name,
                                       data_file_name, file_type, banner);
  LPD_RECEIVE_CONTROL_FILE(strlen(cont_file),cont_file_name);
  if (write(lpd_socket,query,strlen(query))>0)
   {
    retcod=read_lpd_ack(lpd_socket);
    if (_prints_debug != 0)
      printf("Ack to \"receive ctl\" command: %d\n",retcod);  
    if (retcod!=0)
     {
      LPD_SHUTDOWN;
      STOP_TIMER;
      return(-1);
     }
   }
  else
   {
    fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    STOP_TIMER;
    return(-1);
   }
   
  if (_prints_debug != 0)
    printf("Control file length: %d <%s>\n",strlen(cont_file),cont_file);
  if (write(lpd_socket,cont_file,strlen(cont_file))>0)
   {
    write(lpd_socket,&zero,1);
    retcod=read_lpd_ack(lpd_socket);
    if (_prints_debug != 0)
      printf("Ack to control file delivery: %d\n",retcod);  
    if (retcod!=0)
     {
      LPD_SHUTDOWN;
      STOP_TIMER;
      return(-1);
     }
   }
  else
   {
    fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    STOP_TIMER;
    return(-1);
   }

  LPD_SHUTDOWN;
  STOP_TIMER;
  return(0); 
 }

int lpd_client_get_queue_state(char *lpd_server, char *queue,
                               char *buffer, unsigned int size)
 {
  int lpd_socket;
  char *query=NULL;
  unsigned short query_length;
  int retcod;

  lpd_socket = get_lpd_connection(lpd_server);
  if (lpd_socket<0) return(lpd_socket);

  LPD_SEND_QUEUE_STATE_S(queue);

  if (size>0) buffer[0]='\000';

  if (write(lpd_socket,query,strlen(query))>0)
   {
    retcod=read_lpd_response(lpd_socket,buffer,size);
   }
  else
   {
    if (_prints_debug != 0)
      fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    return(-1);
   }

  LPD_SHUTDOWN;

  return(retcod);
 }

int lpd_client_remove_job(char *lpd_server, char *queue,
                               char *user, unsigned int entry)
 {
  int lpd_socket;
  char *query=NULL;
  unsigned short query_length;
  int retcod=0;

  while ((lpd_socket = get_lpd_connection(lpd_server))<0)
   {
    if (errno!=EADDRINUSE) return(lpd_socket);
    sleep(LPD_WAIT_FOR_SOCKET);
   }

  LPD_REMOVE_JOB(queue,user,entry);

  if (write(lpd_socket,query,strlen(query))<=0)
   {
    fprintf(stderr,"Writing to LPD socket on %s: %s\n",lpd_server,
            strerror(errno));
    LPD_SHUTDOWN;
    return(-1);
   }

  LPD_SHUTDOWN;

  return(retcod);
 }

int get_lpd_connection(char *lpd_server)
 {
  int fd_socket;
  int status,original_status;
  static char *last_resolution=NULL;
  static struct addrinfo *ai_ans=NULL, *last_ans=NULL;
  struct addrinfo ai_req, *cur_ans;
  struct addrinfo or_req, *or_ans;
  int cur_or_port;
  char port_string[16];
  fd_set writefs;
  struct timeval timeout;
  int retcod;

  /* Get destination address */

  if (last_resolution == NULL || (strcmp(last_resolution, lpd_server) != 0))
   {
    if (last_resolution != NULL)
     {
      free(last_resolution);
      freeaddrinfo(ai_ans);
      last_resolution = NULL;
      ai_ans = NULL;
     }
    ai_req.ai_flags = 0;
    ai_req.ai_family = PF_UNSPEC;
    ai_req.ai_socktype = SOCK_STREAM;
    ai_req.ai_protocol = 0; /* Any protocol is OK */

    if ((retcod = getaddrinfo(lpd_server, "printer", &ai_req, &ai_ans)) != 0)
     {
      fprintf(stderr, "getaddrinfo call for %s failed. return code = %d\n",
              lpd_server, retcod);
      return(-1);
     }
   }

  if (last_resolution != NULL)
   {
    cur_ans = last_ans;
   }
  else cur_ans = ai_ans;

  /* Search for a working transport */
  for(;cur_ans != NULL; cur_ans=cur_ans->ai_next)
   {
    if ((fd_socket=socket(cur_ans->ai_family, cur_ans->ai_socktype,
                          cur_ans->ai_protocol))==-1) continue;


    /* Look for a working origin port for LPD protocol */
    for (cur_or_port = LPD_TALKER_PORT_MAX; cur_or_port > LPD_TALKER_PORT_MIN;
         cur_or_port-- )
     {
      or_req.ai_flags = AI_PASSIVE;
      or_req.ai_family = cur_ans->ai_family;
      or_req.ai_socktype = cur_ans->ai_socktype;
      or_req.ai_protocol = cur_ans->ai_protocol;
      snprintf(port_string,sizeof(port_string),"%5d",cur_or_port);

      if ((retcod = getaddrinfo(NULL, port_string, &or_req, &or_ans)) != 0)
       {
        fprintf(stderr, "getaddrinfo call for origin address failed. return code = %d\n",
                retcod);
        close(fd_socket);
        fd_socket = -1;
        break; 
       }

      if (or_ans == NULL)
       {
        fprintf(stderr, "getaddrinfo call for origin address returned NULL\n");
        close(fd_socket);
        fd_socket = -1;
        break; 
       }

      if (bind(fd_socket,or_ans->ai_addr,or_ans->ai_addrlen)<0)
       {
        freeaddrinfo(or_ans);
        if (errno==EADDRINUSE) continue;
        perror("Cannot bind origin socket");
        close(fd_socket);
        fd_socket = -1;
        break;
       }
      else 
       {
        freeaddrinfo(or_ans);
        break;
       }
     }

    if (fd_socket < 0) continue;

    /* Make socket asynchronous in order to enforce connect timeout */
  
    if ((original_status=fcntl(fd_socket,F_GETFL,NULL)) == -1)
     {
      perror("Cannot fcntl (GETFL) origin socket");
      close(fd_socket);
      continue;
     }
    status = original_status|FNDELAY;
    if ((fcntl(fd_socket,F_SETFL,status)) == -1)
     {
      perror("Cannot fcntl (SETFL) origin socket");
      close(fd_socket);
      continue;
     }
  
    if (connect(fd_socket,cur_ans->ai_addr,cur_ans->ai_addrlen)==-1)
     {
      if (errno!=EINPROGRESS)
       {
        close(fd_socket);
        continue;
       }
     }
  
    FD_ZERO(&writefs);
    FD_SET(fd_socket,&writefs);
    timeout.tv_sec=LPD_CONNECT_TIMEOUT;
    timeout.tv_usec=0;
    if (select(FD_SETSIZE,(fd_set *) 0,&writefs,(fd_set *) 0,&timeout) <= 0)
     {
      close(fd_socket);
      continue;
     }  
  
    /* restore socket (synchronous) status */
    if ((fcntl(fd_socket,F_SETFL,original_status)) == -1)
     {
      perror("Cannot fcntl (SETFL) origin socket");
      close(fd_socket);
      continue;
     }

    /* Everything's OK. Let's save the contact info. */
    if (last_resolution != NULL) free(last_resolution);
    last_resolution = strdup(lpd_server);
    if (last_resolution != NULL)
     {
      last_ans = cur_ans;
     }
    return(fd_socket);
   }

  freeaddrinfo(ai_ans);
  ai_ans = NULL;
  last_resolution = NULL;
  return(-1);
 }

int read_lpd_response(int fd, char *buffer, unsigned int size)
 {
  int transf_bytes; 
  int result_bytes;
  char *data_pointer;
  fd_set sel_set;
  struct timeval timeout;

  timeout.tv_sec = LPD_STATUS_TIMEOUT;
  timeout.tv_usec = 0;

  data_pointer = buffer;
  transf_bytes = size-1;

  while(transf_bytes>0)
   {
    FD_ZERO(&sel_set);
    FD_SET(fd,&sel_set);
    if (select(FD_SETSIZE,&sel_set,NULL,NULL,&timeout)==0) 
      return(size-transf_bytes-1);
       
    result_bytes = read(fd,data_pointer,transf_bytes); 
    if (result_bytes < 0) 
      if (errno==ECONNRESET) result_bytes=0;
      else
       {
        perror("Reading from socket");
        return(result_bytes);
       }
    
    transf_bytes-=result_bytes; 
    data_pointer+=result_bytes;
    *data_pointer = '\000';

    if (result_bytes == 0) break;
   }

  return(size-transf_bytes-1);
 }

int read_lpd_ack(int fd)
 {
  unsigned char retcod;
  int result_bytes;

  result_bytes = read(fd,&retcod,sizeof(retcod));
  if (result_bytes <= 0) return(-1);

  return((int)retcod);
 }

char *compose_lpd_control_file(char *host, char *user, char *name,
                               char *serv_name, int file_type, int banner)
 {
  static char *buffer=NULL;
  int length;
  struct stat file_char;

#ifndef PRINTMI_GENERATED_BANNER
  if (banner)
   {
    length = 2*strlen(host) + 2*strlen(name) + 2*strlen(user) +
             2*strlen(serv_name) + 16;
   }
  else
#endif
   {
    length = strlen(host) + strlen(name) + strlen(user) + 2*strlen(serv_name)
           + 16;
   }

  buffer = (char *)realloc(buffer,length);

  if (buffer==NULL)
   {
    fprintf(stderr,"Out of memory: cannot alloc.\n");
    return(NULL);
   }

  /* Now send control file entries */
  /* Host name, file name, user name, file name */
#ifndef PRINTMI_GENERATED_BANNER
  if (banner)
   {
    if (LPD_IS(LPD_POSTSCRIPT_TYPE, file_type)) 
      sprintf(buffer,"C%s\nH%s\nJ%s\nP%s\nL%s\nl%s\nU%s\nN%s\n\000",
              host,host,name,user,user,serv_name,serv_name,name);
    else
      sprintf(buffer,"C%s\nH%s\nJ%s\nP%s\nL%s\nf%s\nU%s\nN%s\n\000",
              host,host,name,user,user,serv_name,serv_name,name);
   }
  else
#endif
   {
    if (LPD_IS(LPD_POSTSCRIPT_TYPE, file_type))
      sprintf(buffer,"H%s\nP%s\nl%s\nU%s\nN%s\n\000",
              host,user,serv_name,serv_name,name);
    else
      sprintf(buffer,"H%s\nP%s\nf%s\nU%s\nN%s\n\000",
              host,user,serv_name,serv_name,name);
   }

  return(buffer);
 }

void lpd_client_hard_timeout()
 {
  if (watchdog_jump_initialized) longjmp(watchdog_env,1);
  exit(2);
 }
