/*
 * File: printd.c
 *
 * Description:  
 *
 * Author: Stefania Alborghetti - INFN Milan
 * e-mail: stefania.alborghetti@mi.infn.it
 *
 * History revision:
 * 12 Mar 1998 (FP) Handle SIGHUP (will cause queue reload) and SIGUSR1
 *                  (will toggle the debug state)
 * 30 Apr 1998 (FP) Added printd_unlink_temp_file to remove temporary
 *                  file when a copy child process is killed and the
 *                  filter is active.
 *  3 Jun 1998 (FP) Added filename to exec_format calling sequence
 *                  so that it can be printed in the mpage header.
 * 15 Jul 1998 (FP) Added lines-per-page and page-range options, as
 *                  requested by users.
 * 10 Sep 1998 (FP) Added call to optionally add a banner to the file
 *                  just before it is (optionally as well) filtered with GS.
 *                  Also, added options to avoid setting duplex/tray/paper size
 *                  choices on printers which cannot handle those.
 * 11 Sep 1998 (FP) Added cleanup atexit call for file processing child, so
 *                  that it is able to terminate its own file processing
 *                  children.
 * 16 Sep 1998 (FP) Add handling of lpc commands from client.
 *  5 Nov 1998 (FP) Added handling of extra switches for queue aliases.
 *  9 Nov 1998 (FP) Added F***ing MediaType directive for selecting output
 *                  trays on a F***ing Xerox printer.
 * 11 Mar 1999 (FP) Added for() on remote connect select in order to
 *                  process a maximum of PRINTS_MAX_REQUESTS pending commands 
 *                  before going on to the service cycle.
 * 24 Mar 1999 (FP) Bug fix: statusdict formatting options were ignored when
 *                  no other options were present.
 *  1 Jul 1999 (FP) Bug fix: file should NOT be touched if -@ option
 *                  was received from printif.
 * 29 Sep 2000 (FP) Added MediaPosition directive to accommodate new
 *                  Tektronix/Xerox color printer.
 * 13 Jun 2002 (FP) Replaced communication streams into "_wrap" routines
 *                  with poor man's dynamic strings.
 * 27 Sep 2002 (FP) Corrected a bug in the composition of the temp filename
 *                  for psnup and mpage.
 * 23 Jan 2004 (FP) Added support for extra arguments to GS.
 * 25 Mar 2004 (FP) Moved to revised version of 'itostr'.
 *  2 Apr 2004 (FP) Fixed a couple of issues with PDF handling (should
 *                  not convert CR to LF and should not pipe ps2pdf input).
 * 12 Jun 2006 (FP) Use option SPOOLER_QUEUE_NO_LEVEL1 to send
 *                  setpagedevice instead of satusdict (Level 1) directives.
 * 14 Jun 2006 (FP) Added SO_REUSEADDR flag to listen socket to
 *                  allow for faster server restart.
 * 27 Nov 2007 (FP) Added optional PS-Level2 'Mediatype' directive.
 *                  This is checked only when 'nolevel1' is true.
 *  4 Aug 2009 (FP) Added IPv6 support via getaddrinfo().
 * 12 Mar 2010 (FP) Add option to reject files whose 'file' info contains
 *                  a given string. 
 * 20 Jul 2010 (FP) Added gs_output_filter option to support a few
 *                  Epson paperweights.
 * 24 Jan 2017 (FP) Pass paper size option to ghostscript filter. Use A4 by
 *                  default.
 *
 */

#include     "printd.h"

/* FIXME: should include all the prototypes in spooler_queue.h */
/*        -name conflicts (e.g. 'remove_job' need to be removed- */
char  *media_type(char *queue_name);
char  *gs_output_filter(char *queue_name);

struct control param;
int _prints_debug=0;
int _prints_primary=(-1);
int _prints_need_to_reload_queues = FALSE;
char *_prints_other_server = NULL;
int _prints_parent_pid;
char _prints_temp_file_to_cleanup[FILENAME_MAX];
int  _prints_subprocess_to_cleanup=0;
char *_prints_page_range = NULL;

#define PRINTS_PAGE_DIRECTIVE "%%Page:"
#define PRINTS_BOGUS_PAGE_DIRECTIVE "%Page:"
#define PRINTS_ENDPROLOG_DIRECTIVE "%%EndProlog"
#define PRINTS_BOF_DIRECTIVE "%!"
#define PRINTS_END_DIRECTIVE "end"
#define PRINTS_ENDCOMMENTS_DIRECTIVE "%%EndComments"
#define PRINTS_CREATOR_DIRECTIVE "%%Creator:"

#define PRINTS_MAX_LISTEN_SOCKETS 16

main(int argc, char *argv[])
{
   int option,n;  
   int sockfd[PRINTS_MAX_LISTEN_SOCKETS], newsockfd, clilen;
   int n_listen_sockfd=0, n_fd;
   int n_pending;
   int n_proc_reqs;
   fd_set readfs;
   struct timeval timeout;
   char resolved_client_name[NI_MAXHOST];
   char resolved_client_num[NI_MAXHOST];
   struct sockaddr_storage tentative_client_addr;
   int real_cli_addr_size;
   struct sockaddr *cli_addr;
   struct addrinfo ai_req, *ai_ans, *cur_ans;
   char port_string[16];
   struct passwd *regular_user;
   allowed_host *current_host;
   char *pname;
   time_t now;
   char time_str[80];
   char *opt_str="s:p:dV";
   FILE *pid_file;
   const int socket_reuse_flag = 1;
   
   allowed_host *allowed_host_head = NULL;

   signal(SIGPIPE,SIG_IGN);
   signal(SIGHUP, printd_flag_queue_reload);
   signal(SIGUSR1,printd_toggle_debug);
   _prints_parent_pid = getpid();
   atexit(printd_exit_cleanup);
    
   /* Let's switch to a harmless effective UID */
   regular_user=getpwnam(PRINTD_REGULAR_USER);
   if (regular_user!=NULL)
    {
     setregid(-1,regular_user->pw_gid);
     setreuid(-1,regular_user->pw_uid);
    }
    
   pname = argv[0];

   /*Initialize global structure*/
   param.copies=1;
   param.reserve=PRINTMI_RESERVE_NOP;
   param.id_only=0;
   param.lpc_message=NULL;
   *param.file='\000';
   *param.user='\000';
   *param.host='\000';
   *param.printer='\000';
   *param.entry='\000';
   
   while ((option = getopt(argc, argv, opt_str)) != -1) {
      switch(option){
         
	 case 'd':
               _prints_debug=1;  /*activated debug*/
               printf("Activating debug option...\n");
	       break;
	 
	 case 'p':
	      if(_prints_primary != -1) {
#define PRINTD_USAGE "Usage: %s [-d] {-s primary_server_addr | -p secondary_server_addr}\n"
	      fprintf(stderr,PRINTD_USAGE, argv[0]);
	      exit(-1);
	      }  
	      _prints_primary=1; /*we are being executed as the primary*/
	      if(_prints_debug !=0) 
	         printf("Optarg is: %s \n",optarg);
	      printf("Running as primary server.\n");
              if (does_resolve(optarg) == 0) {
	        fprintf(stderr,"%s: -p argument (%s) does not resolve to a valid host name. Exiting.", argv[0], optarg);
                exit(-1);
              }
	      _prints_other_server=optarg;
	      break;
	 
	 case 's':
	      if(_prints_primary != -1) {
	        fprintf(stderr,PRINTD_USAGE, argv[0]);
	        exit(-1);
	      }  
	      _prints_primary=0; /*we are being executed as the secondary*/
	      if(_prints_debug !=0) 
	         printf("Optarg is: %s \n",optarg); 
              if (does_resolve(optarg) == 0) {
	        fprintf(stderr,"%s: -s argument (%s) does not resolve to a valid host name. Exiting.", argv[0], optarg);
                exit(-1);
              }
	      printf("Running as secondary server.\n");
	      _prints_other_server=optarg;
	      break; 
	 case 'V':
	 case 'v':
            printf("Version: %s \n",SERVER_VERSION);
            exit(0);       
	 default:
	    fprintf(stderr,PRINTD_USAGE,argv[0]);
	    exit(-1);
      }      
   }
   
   if (_prints_primary == -1) {
       _prints_primary=1; /*we are being executed as the primary*/
       printf("Running as primary server (no secondary).\n");
       _prints_other_server=NULL;
   }	      

   snprintf(port_string,sizeof(port_string),"%5d",SERV_TCP_PORT);

   ai_req.ai_flags = AI_PASSIVE;
   ai_req.ai_family = PF_UNSPEC;
   ai_req.ai_socktype = SOCK_STREAM;
   ai_req.ai_protocol = 0; /* Any protocol is OK */

   if (getaddrinfo(NULL, port_string, &ai_req, &ai_ans) != 0)
        err_dump("printd: cannot get server address");

   /* Find all addresses we can bind on */
   for (cur_ans = ai_ans; cur_ans != NULL; cur_ans = cur_ans->ai_next) {
      sockfd[n_listen_sockfd] = socket(cur_ans->ai_family, 
                                       cur_ans->ai_socktype,
                                       cur_ans->ai_protocol);
      if (sockfd[n_listen_sockfd] < 0) continue;
   
      /* Set option to reuse listen address for faster server restart */
      setsockopt(sockfd[n_listen_sockfd],SOL_SOCKET,SO_REUSEADDR,&socket_reuse_flag,sizeof(socket_reuse_flag));
   
      if (bind(sockfd[n_listen_sockfd],
               cur_ans->ai_addr, cur_ans->ai_addrlen) < 0) {
         close(sockfd[n_listen_sockfd]);
         continue;
      }
      if (listen(sockfd[n_listen_sockfd], 5) < 0) {
         close(sockfd[n_listen_sockfd]);
         continue;
      }
      /* So far, so good. */
      n_listen_sockfd++;
   }
            
   freeaddrinfo(ai_ans);

   if (n_listen_sockfd <= 0) {
      err_dump("printd: cannot listen on any socket");
   }

   /* Initialize print spooler queues  & host access from config file(s) */
   initialize_queues(SPOOL_SQUEUES_DB); 
   initialize_host_access(&allowed_host_head); 
   
   /* Write PID into ".pid" file because Marcello likes it */
   if ((pid_file = fopen(PRINTD_PID_FILE,"w"))!=NULL)
    {
     fprintf(pid_file,"%d\n",getpid());
     fclose(pid_file);
    }
   
   if(_prints_debug == 1){
      printf("Listening...\n");
   }     
   for(; ;) {
      FD_ZERO(&readfs);
      for (n_fd=0; n_fd<n_listen_sockfd; n_fd++) {
         if (sockfd[n_fd] >= 0) FD_SET(sockfd[n_fd],&readfs);
      }
      timeout.tv_sec=TIMEOUT;  
      timeout.tv_usec=0;   
      for (n_proc_reqs = 0;
           n_proc_reqs < PRINTS_MAX_REQUESTS &&
           (n_pending = select(FD_SETSIZE,&readfs,(fd_set *) 0,(fd_set *)
                  0,&timeout)) >= 1;
           n_proc_reqs+=n_pending) { /*connection present*/

         for(n_fd = 0; n_fd<n_listen_sockfd; n_fd++) {
           if (sockfd[n_fd] < 0) continue;
           if (!(FD_ISSET(sockfd[n_fd],&readfs))) continue;

           real_cli_addr_size = sizeof(tentative_client_addr);
           cli_addr = (struct sockaddr *)(&tentative_client_addr);
   
           if((newsockfd = accept(sockfd[n_fd], cli_addr, &real_cli_addr_size)) == -1) 
              err_sys("server: accept error");               
           
           strcpy(resolved_client_num, "-unknown-");
           /* Always fill in numeric string if possible */
           getnameinfo(cli_addr, real_cli_addr_size,
                       resolved_client_num, sizeof(resolved_client_num),
                       NULL, 0, NI_NUMERICHOST);
           
           /* Check whether address can be resolved to a name */
           if (getnameinfo(cli_addr, real_cli_addr_size,
                           resolved_client_name, sizeof(resolved_client_name),
                           NULL, 0, NI_NAMEREQD) < 0) {
             now = time(0);
             strncpy(time_str,ctime(&now),sizeof(time_str));
             time_str[sizeof(time_str)-1]='\000';
             time_str[strlen(time_str)-1]='\000';
   
             fprintf(stderr,
                     "%s: refused connection from unresolvable address %s\n",
                     time_str,resolved_client_num);
             shutdown(newsockfd,SHUT_RDWR);
             close(newsockfd);
             continue;
           }
             
           if (allowed_host_head != NULL)
            {
             for(current_host = allowed_host_head; current_host != NULL;
                 current_host = current_host->next)
              {
               if (strstr(resolved_client_name,current_host->name)!=NULL) 
                break;
              }
             if (current_host == NULL)
              {
               if (_prints_debug != 0)
                {
                 now = time(0);
                 strncpy(time_str,ctime(&now),sizeof(time_str));
                 time_str[sizeof(time_str)-1]='\000';
                 time_str[strlen(time_str)-1]='\000';
   
                 fprintf(stderr,"%s: denied access from host <%s>\n",
                         time_str, resolved_client_name);
                }
               shutdown(newsockfd,SHUT_RDWR);
               close(newsockfd);
               continue;
              }
             if (_prints_debug != 0)
               printf("%s == %s BINGOOO!\n",current_host->name,resolved_client_name);
            }
   
           if(_prints_debug) {
              now = time(0);
              strncpy(time_str,ctime(&now),sizeof(time_str));
              time_str[sizeof(time_str)-1]='\000';
              time_str[strlen(time_str)-1]='\000';
              printf("%s: connection from %s (%s)\n",time_str,
                     resolved_client_name,resolved_client_num);
           } 
           str_serv(newsockfd);              
   	   close(newsockfd);     
       }
    }   
     if (_prints_need_to_reload_queues) 
      {
       spooler_queue_reload();
       initialize_host_access(&allowed_host_head);
       _prints_need_to_reload_queues = FALSE;
       if (_prints_debug!=0)
        {
         printf("---> Reloaded queue & access database.\n");
         fflush(stdout);
        }
      }
     do_print_wrap();
     fflush(stdout);
   } 
}                              
                                

/*
 * str_serv
 */  
  
int str_serv(int sockfd)
{
  char read_line[MAXLINE]; 
  int n;
 
  /*Communicates version numbers*/
  n=strlen(CLIENT_VERSION);
  if (write(sockfd, CLIENT_VERSION, n) != n){
     if(errno == EPIPE) 
       return(0);
     else {
       perror("str_serv: socket write error");
       return (-1);
     }
  } 
  if (write(sockfd, " / ", 3) != 3){
     if(errno == EPIPE)
        return(0);
     else {
        perror("str_serv: socket write error");
        return (-1);
     }
  }

  n=strlen(SERVER_VERSION);
  if (write(sockfd, SERVER_VERSION, n) != n){
     if(errno == EPIPE)
       return(0);
     else {
       perror("str_serv: socket write error");
       return (-1);
     }
  }

  if (write(sockfd, "\n", 1) != 1){
     if(errno == EPIPE) 
        return(0);
     else {
        perror("str_serv: socket write error");
        return (-1);
     }
  } 
  
  
  /*Read request type */
  if ((n=readline(sockfd, read_line , MAXLINE)) <= 0)
       return -1;
  if((n <= MAXLINE)&&(n>=1)) read_line[n-1]='\000';     
  if (_prints_debug != 0)
     printf("Output from client: <%s>\n",read_line);
  if (strcmp(read_line,"lpr") == 0){
      print_file(sockfd); 
      return 1;   
  }else if (strcmp(read_line,"lpq") == 0){
      show_queue(sockfd); 
      return 1;   
  }else if (strcmp(read_line,"lprm") == 0){
      remove_job(sockfd); 
      return 1;
  }else if (strcmp(read_line,"lpc") == 0){
      control_queue(sockfd); 
      return 1;
  }else if (strcmp(read_line,"vers") == 0){
      send_version(sockfd); 
      return 1;       
  }else if (strcmp(read_line,"n_jobs") == 0){
      get_n_queued_jobs(sockfd);
  }else if (strstr(read_line,"+++") == 0){
      return 0;        
  }else {
      printd_error(sockfd);
      return 0; 
  }                                             
}

/*
 * get_n_queued_jobs
 */

int get_n_queued_jobs(int sockfd)
{
   int n,n_jobs;
   
   n=isempty_wrap();
   n_jobs=htonl(n);
   fflush(stdout);
   write(sockfd,&n_jobs,sizeof(int));
   close(sockfd);
   return 0;   

}


/*
 * print_file
 */

int print_file(int sockfd)
{
    char   read_line[MAXLINE];
    char   send_line[MAXLINE]; 
    char*  format=NULL;
    char*  gs_driver;
    char*  buf;
    char   c;
    char*  spool_item;
    char*  readp;
    char*  writep;
    char   c_one='\001';
    int         fp;
    int         i,n,len; 
    static unsigned int spool_number =1;
    int         line_length, length;
    int         readl;
    int         first_line=TRUE;
    int         allow_formatting=TRUE;
    int         received_file_is_pdf=FALSE;
    char *check_string;
    char *file,*file_start;
    char *unaliased_queue;
    char *extra_options = NULL;
    char **gs_extra_arguments;
    int    gs_extra_arguments_n;
    DSTRING pip;
    char* msg;
    pid_t childpd; 
    char *paper_size = NULL;
    char *pstart, *pend;
    
    mode_t mode=0600; 
    /*  
      *format='\000';         
      *output='\000';
    */

    /* Read control parameters */
    if ((n = readline(sockfd, read_line , FILENAME_MAX)) <=0){
       err_ret("print_file:readline failed");
       n=2;
    }   
    len=sizeof(param.file);
    strncpy(param.file,read_line,len);                 /*filename*/
    if (n <= len) param.file[n-1]='\000';
    else param.file[len-1]='\000';
    /* printf("%s",param.file); */
    if((n = readline(sockfd, read_line , MAXENTRY)) <=0){
      err_ret("print_file:readline failed");
       n=2;
    }   
    len=sizeof(param.user);
    strncpy(param.user,read_line,len);                /* param.user */ 
    if (n <= len) param.user[n-1]='\000';
    else param.user[len-1]='\000';
    /* printf("%s",param.user); */
    if ((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("print_file:readline failed");
       n=2;
    }   
    len=sizeof(param.host);
    strncpy(param.host,read_line,len);               /* param.host */
    if (n <= len) param.host[n-1]='\000';
    else param.host[len-1]='\000';
    /* printf("%s",param.host); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("print_file:readline failed");
       n=2;
    }   
    len=sizeof(param.printer);
    strncpy(param.printer,read_line,len);           /* param.printer */
    if (n <= len) param.printer[n-1]='\000';
    else param.printer[len-1]='\000';
    unaliased_queue = dealias_queue_name(param.printer);
    if (unaliased_queue != param.printer)
     {
      extra_options = get_alias_switches(param.printer);
      strncpy(param.printer,unaliased_queue,sizeof(param.printer));
      if (strlen(unaliased_queue)>=sizeof(param.printer))
       param.printer[sizeof(param.printer)-1]='\000';
     }
    if(_prints_debug !=0)
       printf("Printer is <%s>\n",param.printer); 
        
    if ((n=readint(sockfd,&param.id_only)) == -1)          /*identity only */ 
          err_ret("print_file: read error");
   
    if ((n=readint(sockfd,(int *)&param.uid)) == -1)        /*user uid */ 
          err_ret("print_file: read error");
    
    if ((n=readint(sockfd,&param.reserve)) == -1)          /*reservation */ 
          err_ret("print_file: read error");
     
    if ((n=readint(sockfd,(int *)&param.copies)) == -1)   /* number of copies*/ 
          err_ret("print_file: read error");
    
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("print_file:readline failed");
       n=2;
    }   
    len=sizeof(param.entry);
    strncpy(param.entry,read_line,n-1);                /* param.entry */
    if (n <= len) param.entry[n-1]='\000';
    else param.entry[len-1]='\000';
    if (_prints_debug!=0)
      printf("param.entry = <%s>\n",param.entry); 
    
    spool_item=itostr(spool_number);
    if (spool_item == NULL) spool_item = strdup("NOMEM"); /* Out of memory */
    if(spool_number < 99999) spool_number++;
      else spool_number=1;
      
    len=MAXWORD+2*MAXENTRY+6;
    buf=(char *) malloc(len);
    strncpy(buf,param.user,len);
    len -=strlen(param.user);
    if (len > 1) strncat(buf,"@",len-1);
    len--;
    if (len > 1 ) strncat(buf,param.host,len-1);
    len -=strlen(param.host);
    if (len > 1 ) strncat(buf,":",1);
    len--;

    file = (char *)malloc(strlen(param.file)+1);
    if (file != NULL)
     {
      strcpy(file,param.file);
      file_start = file;
      if((file_start=strrchr(file,']')) !=NULL)
       {
        file_start++;
        for(n=0;n<=strlen(file_start);n++)
         if(file_start[n]==';')
          {
              file_start[n]='\000';
              break;
          }
       }
      else
       {
        file_start = file;
        if((file_start=strrchr(file,'/')) !=NULL)
         {
           file_start++;
         } else {
           file_start=file;
         }
       }
     }

    if (len > 1 ) strncat(buf,file_start,len-1);
    len -=strlen(file_start);
    free(file);
    if (len > 1 ) strncat(buf,".",len-1);
    len--;
    if (len > 1 ) strncat(buf,spool_item,len-1);
     
     if(_prints_debug != 0)
       printf("%s\n",buf);
    
    if (strlen(param.entry)==0) {  /*need to transfer and format file*/
      if((n=readline(sockfd, read_line , MAXLINE))<=0){
         err_ret("print_file:readline failed");
         n=2;
      } 
      if (extra_options != NULL) n+=(strlen(extra_options)+1);
      if (n>2) {
         format =(char*)malloc(n);
         if (format == NULL) err_ret("print_file:out of memory");
         strncpy(format,read_line,n-1);
         format[n-1]='\000'; 
         if (extra_options != NULL) 
          {
           strcat(format," ");
           strcat(format,extra_options);
          }
      } else {
         if(_prints_debug != 0) printf("Null Format string\n");
	 format =(char *)strdup("");
         if (format == NULL) err_ret("print_file:out of memory");
      }	  
       	 					    
        /* The file should NOT be touched if -@ (printif option)
           was received */
      if (strstr(format,"-@")!=NULL) allow_formatting = FALSE;

        /* Extract paper size option for GS */
      if ((pstart = strstr(format,"-P"))!=NULL){
        pstart += 2;
        pend = pstart;
        while (((*pend) >= '-') && ((*pend)<='z')) pend++;
        if (pend > pstart) {
          paper_size = (char *)calloc(1, (int)(pend - pstart) + 1);
          if (paper_size != NULL) {
            memcpy(paper_size, pstart, (int)(pend - pstart));
          } /* If malloc fails will use the default paper format */
        }
      }

      if(_prints_debug != 0){
        printf("Read_line: %s\n",read_line);
        printf("Format string: %s\n",format);
      }
    
      if((childpd = fork()) < 0 )
        err_ret("print_file: Cannot fork");
      else if (childpd == 0) {
    
        if ((fp = creat(buf,mode)) == -1 )
          err_ret("print_file: cannot CREATE print file");
    
        length=0;
        while ((n=readint(sockfd, &line_length)) > 0){
          if (line_length > MAXLINE)
           {
            /* The client cannot send lengths greater than MAXLINE
               under ordinary circumstances. If this happens, we
               have some sort of corruption in the message stream, we
               print a message, destroy the spool file and choke. */
            close(fp);
            unlink(buf);
            err_quit("print_file: received line_length = %d",line_length);
           }
	  length += line_length;
	  /*
	   if (_prints_debug !=0)
	     printf("Line length is: %d Total: %d\n",line_length,length); 
	  */
	  if (line_length <= 0) break; /*EOF*/
	  readp=read_line;
	  readl=line_length;
	  while ((n=read(sockfd,readp,readl)) >0){
            readp+=n;
	    readl-=n;
	    if(readl<=0) break;
	  }
	  if (n < 0) err_ret("print_file: read error");

          writep = read_line;
	 	 
	  if(first_line){
	   first_line = FALSE;
           /* Remove leading garbage from PS file if any. */
	   if ((writep=strstr(read_line,"%!PS-Adobe")) == NULL)
	      writep = read_line;
	   else line_length -= (int)(writep-read_line);
           if (strncmp(writep,"%PDF",4)==0) received_file_is_pdf = TRUE;
          }

          /* Convert CRs to LFs */
          if (allow_formatting && (!received_file_is_pdf))
            for(i=0;i<line_length;i++)
	      if(writep[i] == '\r') writep[i]='\n';

	  if ((i= write(fp,writep,line_length)) == -1)
	     err_sys("print_file:write error");
        }    
        close(fp);
        if ((check_string=check_file(param.printer))!=NULL)
         {
          if (exec_check_file(buf,check_string)==FALSE)
           {
            unlink(buf);
            exit(1); /*Disappear... The server will notice and dequeue the job*/
           }
         }

        if ((exec_format(buf,format)) == -1)
           err_sys("print_file: exec_format failed"); 
        if (format != NULL) free(format);	  
        format = NULL;

        if (_prints_page_range!=NULL && allow_formatting){
          if(_prints_debug !=0)
            printf("print_file:Executing psselect...\n");
          exec_psselect(buf,_prints_page_range); 
          free(_prints_page_range);
          _prints_page_range = NULL;
        }
     
#ifdef PRINTMI_GENERATED_BANNER
        if (wants_banner(param.printer) && allow_formatting){
          if(_prints_debug !=0)
            printf("print_file:Adding banner page...\n");
          exec_add_banner(&param,buf);
        }
#endif

        if ((gs_driver=use_gs(param.printer))!=NULL){
          if (_prints_check_data_file(buf,1)==0) /* PostScript */
           {
            if(_prints_debug !=0)
	      printf("print_file:Executing ghostscript...\n"); 
            gs_extra_arguments = extra_gs_arguments(param.printer, &gs_extra_arguments_n);
            exec_gs(buf,gs_driver,
                    gs_extra_arguments,
                    gs_extra_arguments_n,
                    gs_output_filter(param.printer),
                    paper_size); /*Don't touch gs_driver*/ 
           }
        }	 
        exit(0); /*Child exits*/
     }
     pip=lpr_wrap(&param,buf,childpd);
   }else /*param.entry != 0 i.e. it's a released reservation*/
     pip=lpr_wrap(&param,buf,0);
     
    if (pip != NULL)
     {
      msg = DSTRING_PTR(pip);
      n = strlen(msg);
      if (writen(sockfd, msg, n) != n)
         err_ret("print_file: writen error");
      free(pip);
     }

/* Tell client to exit, no need to wait for long formatting
    (pdf2ps or gs) to end - cannot shutdown socket connection here
    because cannot be sure file has already been transferred comp. */
          
    if ((write(sockfd,&c_one,1))==-1)
       err_ret("print_file:final write error\n");
    if (format != NULL) free(format);
    if (spool_item != NULL) free(spool_item);
    free(buf);                    
    if (paper_size != NULL) free(paper_size);
    return 1;                 
}

/*
 * show_queue
 */


int show_queue(int sockfd)
{
    char read_line[MAXLINE];
    char send_line[MAXLINE]; 
    DSTRING pip;  
    char* msg;
    int n;                
     
    if((n = readline(sockfd, read_line , FILENAME_MAX))<=0){
        err_ret("show_queue:readline failed");
        n=2;
    }   
    strncpy(param.file,read_line,n-1);                    /*filename*/
    param.file[n-1]='\000';
    /* printf("%s",param.file); */   
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("show_queue:readline failed");
       n=2;
    }   
    strncpy(param.user,read_line,n-1);                    /* param.user */
    param.user[n-1]='\000';   
    /* printf("%s",param.user); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("show_queue:readline failed");
       n=2;
    }   
    strncpy(param.host,read_line,n-1);                    /* param.host */
    param.host[n-1]='\000';
    /* printf("%s",param.host); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("show_queue:readline failed");
       n=2;
    }       
    strncpy(param.printer,read_line,n-1);                 /* param.printer */
    param.printer[n-1]='\000';
    if(_prints_debug !=0)
       printf("Printer is <%s>\n",param.printer); 
   
    
    if (strcasecmp(param.printer,"all") == 0)
        *param.printer='\000';      
    
    if ((n=readint(sockfd,&param.id_only)) == -1)          /*identity only */ 
          err_ret("print_file: read error");
    
    if ((n=readint(sockfd,(int *)&param.uid)) == -1)
          err_ret("show_queue: read error");
    
    pip=lpq_wrap(&param);                
    if (pip != NULL)
     {
      msg = DSTRING_PTR(pip);
      if(_prints_debug != 0) 
        printf("%s",msg);   
      n = strlen(msg);
      if (writen(sockfd, msg, n) != n)
         err_ret("print_file: writen error");
      free(pip);
     }
      
   return 1;                 
}

/*
 * control_queue
 */


int control_queue(int sockfd)
{
    char read_line[MAXLINE];
    char send_line[MAXLINE];
    DSTRING pip;
    char* msg;
    int n;

    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("show_queue:readline failed");
       n=2;
    }
    strncpy(param.user,read_line,n-1);                    /* param.user */
    param.user[n-1]='\000';
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("control_queue:readline failed");
       n=2;
    }
    strncpy(param.host,read_line,n-1);                    /* param.host */
    param.host[n-1]='\000';
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("control_queue:readline failed");
       n=2;
    }
    strncpy(param.printer,read_line,n-1);                 /* param.printer */
    param.printer[n-1]='\000';
    if(_prints_debug !=0)
       printf("Printer is <%s>\n",param.printer);
   
    if (strcasecmp(param.printer,"all") == 0)
        *param.printer='\000';

     if ((n=readint(sockfd,&param.id_only)) == -1)          /*identity only */
          err_ret("print_file: read error");

    if ((n=readint(sockfd,(int *)&param.uid)) == -1)
          err_ret("show_queue: read error");

    if ((n=readint(sockfd,(int *)&param.reserve)) == -1)
          err_ret("show_queue: read error");

    if((n = readline(sockfd, read_line, MAX_STATUS_MESSAGE_LEN))<=0){
       err_ret("control_queue:readline failed");
       n=2;
    }
    param.lpc_message = (char *)malloc(n);
    if (param.lpc_message != NULL) 
     {
      strncpy(param.lpc_message,read_line,n-1); 
      param.lpc_message[n-1]='\000';
     }

    pip=lpc_wrap(&param);

    free (param.lpc_message);
    param.lpc_message = NULL;

    if (pip != NULL)
     {
      msg = DSTRING_PTR(pip);
      n = strlen(msg);
      if(_prints_debug != 0)
        printf("%s",msg);
      if (writen(sockfd, msg, n) != n)
         err_ret("print_file: writen error");
      free(pip);
     }

   return 1;
}

/*
 * send_version
 */

int send_version(int sockfd)
{               
  int n;
  char host_name[MAXENTRY];
  
  if((n = readline(sockfd, host_name , MAXENTRY))<=0){
        err_ret("send_version:readline failed");
        n=2;
    }   
    
    if (n <= MAXENTRY) host_name[n-1]='\000';
  
  if(_prints_debug !=0)
    printf("send_version: comunicating version number to client <%s>\n",
           host_name);
  
  n=strlen(SERVER_VERSION);
  
  if (write(sockfd, SERVER_VERSION, n) != n){
      printf("send_version: socket write error");
       return (-1);
  }  
  if (write(sockfd, "\n", 1) != 1){
      printf("str_serv: socket write error");
       return (-1);
  }  
   
  return 1;                 
}


/*
 * remove_job
 */

int remove_job(int sockfd)
{
    char read_line[MAXLINE];
    char send_line[MAXLINE]; 
    int n;
    DSTRING pip; 
    char* msg;
    
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("remove_job:readline failed");
       n=2;
    }   
    strncpy(param.entry,read_line,n-1);                   /*param.entry*/
    param.entry[n-1] = '\000';
    /* printf("%s",param.entry); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("remove_job:readline failed");
       n=2;
    }   
    strncpy(param.user,read_line,n-1);                    /* param.user */  
    param.user[n-1]='\000';
    /* printf("%s",param.user); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("remove_job:readline failed");
       n=2;
    }   
    strncpy(param.host,read_line,n-1);                    /* param.host */
    param.host[n-1]='\000';
    /* printf("%s",param.host); */
    if((n = readline(sockfd, read_line , MAXENTRY))<=0){
       err_ret("remove_job:readline failed");
       n=2;
    }   
    strncpy(param.printer,read_line,n-1);                 /* param.printer */
    param.printer[n-1]='\000';
    /* printf("%s",param.printer); */ 
        
     if ((n=readint(sockfd,&param.id_only)) == -1)          /*identity only */ 
          err_ret("print_file: read error");
    
    if ((n=readint(sockfd,(int *)&param.uid)) == -1)
          err_ret("remove_job: read error");
        
    pip=lprm_wrap(&param);
    if (pip != NULL)
     {
      msg = DSTRING_PTR(pip);
      n = strlen(msg);
      if (writen(sockfd, msg, n) != n)
         err_ret("print_file: writen error");
      free(pip);
     }

    return 1;  
}

/*******************************************************************************/
 
get_time()
{
   struct timeval tp;
   struct timezone tzp;
   
   if((gettimeofday(&tp, &tzp)) == -1)
     err_sys("sync_load: Cannot execute gettimeofday\n");
   /* printf("Temp corrente %i \n",tp.tv_sec); */
   ctime(&tp.tv_sec);   
   return tp.tv_sec;  
}


/****************************************************************************/
int printd_error(int sockfd)
{
   char send_line[MAXLINE];
 
   strncpy(send_line,"Unrecognized request!",sizeof(send_line));
   send_line[sizeof(send_line)-1]='\000';
   write(sockfd,send_line,strlen(send_line));
   return 0;
}       

/*
 * exec_gs 
 */
int exec_gs(char* buf, char* _gs_driver, 
            char **extra_arguments, int n_extra_arguments,
            char *out_filter, char *paper_size)
{

#define EXEC_GS_DEFAULT_ARGS    13
#define GS_DEV_OPT     "-sDEVICE="
#define GS_OUT_OPT     "-sOutputFile="

  
  int child_process,retpid,status,n;
  FILE *input_file;
  FILE* cmdout;
  char* _gs_output=NULL;
  char* _gs_device_opt=NULL;
  char* _gs_output_opt=NULL;
  char **arguments;
  int arg_max;
  int out_filter_len=0;
  char* _gs_paper_size=NULL;
  char *paper_size_format = "-sPAPERSIZE=%s";

  arguments = (char **)malloc((EXEC_GS_DEFAULT_ARGS + n_extra_arguments) * sizeof(char *));
  if (arguments == NULL)
   {
    fprintf(stderr,"Out of Memory\n");
    return(-1);
   }
  
  _gs_output=(char*)malloc(strlen(buf)+5);
  if (_gs_output == NULL)
   {
    fprintf(stderr,"Out of Memory\n");
    free(arguments);
    return(-1);
   }
  strncpy(_gs_output,buf,strlen(buf));
  _gs_output[strlen(buf)]='\000';
  strncat (_gs_output,".#gs",4);
 
  if (out_filter != NULL) out_filter_len = strlen(out_filter)+6;

  _gs_output_opt=(char*)malloc(strlen(_gs_output) +
                               out_filter_len + sizeof(GS_OUT_OPT)+1);
  if (_gs_output_opt == NULL)
   {
    fprintf(stderr,"Out of Memory\n");
    free(arguments);
    free(_gs_output);
    return(-1);
   }

  if (paper_size = NULL) paper_size = "a4";
  _gs_paper_size = (char*)malloc(strlen(paper_size_format) + strlen(paper_size));
  if (_gs_paper_size == NULL)
   {
    fprintf(stderr,"Out of Memory\n");
    free(arguments);
    free(_gs_output);
    free(_gs_output_opt);
    return(-1);
   }
  sprintf(_gs_paper_size, paper_size_format, paper_size);

  if (out_filter == NULL)
   {
    strncpy(_gs_output_opt,GS_OUT_OPT,sizeof(GS_OUT_OPT));
    _gs_output_opt[sizeof(GS_OUT_OPT)]='\000';
    strncat(_gs_output_opt,_gs_output,strlen(_gs_output));
   }
  else
   {
    sprintf(_gs_output_opt,"%s| %s > %s", GS_OUT_OPT, out_filter, _gs_output);
   }

  _gs_device_opt=(char*)malloc(strlen(_gs_driver)+sizeof(GS_DEV_OPT)+1);
  if (_gs_device_opt == NULL)
   {
    fprintf(stderr,"Out of Memory\n");
    free(arguments);
    free(_gs_output);
    free(_gs_output_opt);
    free(_gs_paper_size);
    return(-1);
   }
  strncpy(_gs_device_opt,GS_DEV_OPT,sizeof(GS_DEV_OPT));
  _gs_device_opt[sizeof(GS_DEV_OPT)]='\000';
  strncat(_gs_device_opt,_gs_driver,strlen(_gs_driver));

   if(_prints_debug) {
    printf("exec_gs: Setting gs arguments...\n");
    fflush(stdout);
  } 

  arg_max=0;
  arguments[arg_max++]=PRINT_GHOSTSCRIPT;
  arguments[arg_max++]=_gs_device_opt;
  arguments[arg_max++]=_gs_output_opt;
  arguments[arg_max++]=_gs_paper_size;
  arguments[arg_max++]="-dNOPAUSE";

  if (_prints_debug==0) arguments[arg_max++]="-q";
 
  if (extra_arguments!=NULL)
   {
    for (n=0; n<n_extra_arguments; n++)
     {
      arguments[arg_max++] = extra_arguments[n];
     }
   }

  arguments[arg_max++]="-";
  arguments[arg_max] = NULL;

  n=0;
  if(_prints_debug !=0)
    while((arguments[n]!=NULL)&&(n<arg_max))
     {
       printf("GS arguments[%d] is <%s>.\n",n,arguments[n]);
       n++;
     }

  if ((input_file=fopen(buf,"r"))==NULL)
   {
    fprintf(stderr,"Cannot open file %s: %s\n",buf,strerror(errno));
    free(_gs_output);
    free(_gs_output_opt);
    free(_gs_paper_size);
    free(_gs_device_opt);
    free(arguments);
    return(-1);
   }
  fclose(input_file);   
     
  strncpy(_prints_temp_file_to_cleanup,_gs_output,
           sizeof(_prints_temp_file_to_cleanup));
  atexit(printd_unlink_temp_file);
  if(_prints_debug!=0)
    printf("exec_gs: Beginning to fork...\n");
  if ((child_process = fork())<0)
    err_sys("exec_gs:Cannot fork");
  if (child_process == 0){
     fclose(stdin);
     memcpy(stdin,fopen(buf,"r"),sizeof(FILE));
     exit(execv(PRINT_GHOSTSCRIPT,arguments));    
  }               

  _prints_subprocess_to_cleanup = child_process;

  /* We are in the format child, so we can wait ... */
  if ((retpid=waitpid(child_process, &status, 0)) == -1)
   {
    fprintf(stderr,"Error waiting for PID %d: %s \n",child_process,
                                              strerror(errno));
   }
  
  _prints_subprocess_to_cleanup = 0;

  if(_prints_debug!=0)
    printf("Exec_gs: status is %d\n",status);
  rename(_gs_output,buf);
  
  free(_gs_output);
  free(_gs_output_opt);
  free(_gs_paper_size);
  free(_gs_device_opt);
  free(arguments);
  
  return(0);
}

/*
 * exec_pdf_2_ps 
 */
int exec_pdf_2_ps(char* buf)
{
   
#define EXEC_PDF_2_PS_MAX_ARGS  5
  
  int child_process,retpid,status;
  FILE *input_file;
  char *output=NULL;
  char *arguments[EXEC_PDF_2_PS_MAX_ARGS+2];
  
  output=(char*)realloc(output,strlen(buf)+5);
  strncpy(output,buf,strlen(buf));
  output[strlen(buf)]='\000';
  strncat (output,".#ps",4);

  arguments[0]=PRINT_PDF_2_PS;
  arguments[1]=buf;
  arguments[2]=output;
  arguments[3]=NULL;

  if ((input_file=fopen(buf,"r"))==NULL)
   {
    fprintf(stderr,"exec_pdf_2_ps: cannot open file %s: %s\n",buf,strerror(errno));
    free(output);
    return(-1);
   }
  fclose(input_file);   
     
  strncpy(_prints_temp_file_to_cleanup,output,
           sizeof(_prints_temp_file_to_cleanup));
  atexit(printd_unlink_temp_file);

  if(_prints_debug!=0)
    printf("exec_pdf_2_ps: Beginning to fork...\n");
  if ((child_process = fork())<0)
    err_sys("exec_pdf_2_ps:Cannot fork");
  if (child_process == 0){
     exit(execv(PRINT_PDF_2_PS,arguments));    
  } 

  _prints_subprocess_to_cleanup = child_process;

  /* We are in the format child, so we can wait ... */
  if ((retpid=waitpid(child_process, &status, 0)) == -1)
   {
    fprintf(stderr,"exec_pdf_2_ps: Error waiting for PID %d: %s \n",
                                child_process, strerror(errno));
   }

  _prints_subprocess_to_cleanup = 0;

  printf("Exec_pdf_2_ps: status is %d\n",status);
  rename(output,buf);
  
  free(output);
  return(0);
}  

/* exec_check_file: execute a system "file" command, requiring
   a certain string in the output */

int exec_check_file(char *buf, char *check_string)
 {
  FILE *file_output;
  char command_string[32+FILENAME_MAX];
  char out_buffer[512];
  char *cks;
  int negate = FALSE;
  int len;

  if (check_string[0] == '!' && check_string[1] == '!')
   {
    cks = check_string +2;
    negate = TRUE;
   }
  else cks = check_string;

  sprintf(command_string,"file '%s'",buf);

  file_output = popen(command_string,"r");
  if (file_output != NULL)
   {
    len = fread(out_buffer, sizeof(char), sizeof(out_buffer) - 1 , file_output);
    if (len>0) out_buffer[len]='\000';
    if (strstr(out_buffer,cks)!=NULL) return (negate ? FALSE : TRUE);
    else                              return (negate ? TRUE  : FALSE);
   }
 }

/*
 * exec_psselect
 */
int exec_psselect(char* buf, char* page_range)
{

#define EXEC_PSSELECT_MAX_ARGS   4 

  int child_process,retpid,status,n;
  FILE *input_file;
  char *psselect_output=NULL;
  char *psselect_range=NULL;
  char *arguments[EXEC_PSSELECT_MAX_ARGS+2];
 
  psselect_output=(char*)malloc(strlen(buf)+6);
  strncpy(psselect_output,buf,strlen(buf));
  psselect_output[strlen(buf)]='\000';
  strncat (psselect_output,".#pss",5);

  psselect_range=(char*)malloc(strlen(page_range)+4);
  sprintf(psselect_range,"-p%s",page_range);

  if(_prints_debug) {
    printf("exec_psselect: Setting psselect arguments...\n");
    fflush(stdout);
  }

  arguments[0]=PRINT_PROGRAM_PSSELECT;
  arguments[1]=psselect_range;
  arguments[2]=buf;
  arguments[3]=psselect_output;
  arguments[4]=NULL;

  n=0;
  if(_prints_debug !=0)
    while((arguments[n]!=NULL)&&(n<EXEC_PSSELECT_MAX_ARGS))
     {
       printf("psselect arguments[%d] is <%s>.\n",n,arguments[n]);
       n++;
     }

  if ((input_file=fopen(buf,"r"))==NULL)
   {
    fprintf(stderr,"exec_psselect: cannot open file %s: %s\n",buf,strerror(errno
));
    free(psselect_output);
    free(psselect_range);
    return(-1);
   }
  fclose(input_file);


  strncpy(_prints_temp_file_to_cleanup,psselect_output,
           sizeof(_prints_temp_file_to_cleanup));
  atexit(printd_unlink_temp_file);

  if(_prints_debug!=0)
    printf("exec_psselect: Beginning to fork...\n");
  if ((child_process = fork())<0)
    err_sys("exec_psselect:Cannot fork");
  if (child_process == 0){
     exit(execv(PRINT_PROGRAM_PSSELECT,arguments));
  }

  _prints_subprocess_to_cleanup = child_process;

  /* We are in the format child, so we can wait ... */
  if ((retpid=waitpid(child_process, &status, 0)) == -1)
   {
    fprintf(stderr,"Error waiting for PID %d: %s \n",child_process,
                                              strerror(errno));
   }

  _prints_subprocess_to_cleanup = 0;

  if(_prints_debug!=0)
    printf("Exec_psselect: status is %d\n",status);
  rename(psselect_output,buf);
 
  free(psselect_output);
  free(psselect_range);
 
  return(0);
}

/*
 * exec_add_banner
 */
int exec_add_banner(struct control *param,char* buf)
{
  FILE *input_file;
  FILE *output_file;
  char buffer[2048];
  char *addbanner_output=NULL;
  int nread;

  addbanner_output=(char*)malloc(strlen(buf)+6);
  strncpy(addbanner_output,buf,strlen(buf));
  addbanner_output[strlen(buf)]='\000';
  strncat(addbanner_output,".#adb",5);

  if ((input_file = fopen(buf,"r"))==NULL)
   {
    fprintf(stderr,"Warning: could not open %s for reading:",buf);
    perror("");
    free(addbanner_output);
    return -1;
   }

  if ((output_file = fopen(addbanner_output,"w"))==NULL)
   {
    fprintf(stderr,"Warning: could not create %s for writing:",addbanner_output);
    perror("");
    free(addbanner_output);
    return -1;
   }

  printmi_banner_page (output_file, param->user, param->host, param->printer,
                                    param->file);
  
  while (!feof(input_file))
   {
    if ((nread=fread(buffer,1,sizeof(buffer),input_file))==EOF) break;
    if (fwrite(buffer,1,nread,output_file)==EOF) break;
   }

  fclose(input_file);
  fclose(output_file);

  rename(addbanner_output,buf);

  free(addbanner_output);

  return(0);
}


/*
 * exec_format 
 */

exec_format(char* buf, char* old_format)
{ 
  pid_t child_process,retpid;
  char* format=NULL;
  char* pt_format=NULL;
  FILE* cmdout;
  FILE* spool_file;
  char* spool=NULL;
  int  line_index;
  char line[512];
  char cur, old;
  int status,retcod;
  int _prints_postscript=FALSE;
  int _prints_pdf=FALSE;
  int _prints_ascii=FALSE;
  int _prints_format=FALSE;
  int _prints_set_duplex_mode=FALSE;
  int _prints_force_no_duplex=FALSE;
  int _prints_set_tumble_mode=FALSE;
  int _prints_set_tray=FALSE;
  int _prints_set_nolevel1=FALSE;
  char *_prints_set_media_type=NULL;
  int _prints_manual_feed=FALSE;
  int _prints_transp_tray=FALSE;
  int _prints_media_pos=FALSE;
  int stdict_opt_len=0; /*length of statusdict format options*/
  int tray_num=0; /*default tray number*/
  int _prints_set_opt_done=FALSE;
  int insert_before_marker,got_insert_match;
  int n_ps_insert_retries;
  int cck;
  char *cur_check;
  char *ps_directive_insert_marker[4]=
   { PRINTS_PAGE_DIRECTIVE, PRINTS_BOGUS_PAGE_DIRECTIVE, 
     NULL, NULL
   };
  
#define EXEC_FORMAT_MAX_ARGS 21
  char *arguments[EXEC_FORMAT_MAX_ARGS+2];
  char *line_scan;

  char header_string[60];
  char header_concat[2];
  char argument[132];
  int i,n_args,format_length;
  

#define FORMAT_OFFSET  60 /*Should be calculated properly...*/
  format=(char*)malloc(strlen(old_format)+FORMAT_OFFSET);
  if (format == NULL) {
    fprintf(stderr,"exec_format: Out of memory.\n");
    err_sys("exec_format: exiting");     
  }
      
#define SPOOL_SIZE strlen(buf)+10
  spool=(char*)malloc(SPOOL_SIZE);
  if (spool == NULL) {
    fprintf(stderr,"exec_format: Out of memory.\n");
    err_sys("exec_format: exiting");     
  }
      
  /*
    printf("Hello from exec_format...\n");
    fflush(stdout);
  */
  
  if((retcod=_prints_check_data_file(buf,1))==0) _prints_postscript=TRUE;
    else if(retcod==1) _prints_pdf=TRUE;
    else _prints_ascii=TRUE;
   
  /*
     printf("Format check returns: %d ps: %d pdf: %d asc: %d\n",
         retcod, _prints_postscript, _prints_pdf, _prints_ascii);
     fflush(stdout);
  */
  
  if(_prints_pdf==TRUE){
     if(_prints_debug !=0) 
       printf("Exec_format: file is pdf...calling pdf2ps.\n");
     exec_pdf_2_ps(buf);
     _prints_postscript=TRUE;
  }   
  
/*Modifing format options for psnup or mpage*/ 
 
  if(_prints_postscript==TRUE)
    if ((_prints_client_to_psnup(old_format,format))==0)
       _prints_format=TRUE; /*Return 0 only if null format string or "-x" is present*/
  if(_prints_ascii==TRUE)
    if ((_prints_client_to_mpage(old_format,format))==0)
       _prints_format=TRUE;                                      
  
  if(_prints_debug !=0){
    printf("Exec_format file is: <%s>\n",buf);
    printf("Exec_format format is: <%s>\n",format);
    if(_prints_postscript==TRUE) 
       printf("Exec_format: file is postscript.\n");
    else printf("Exec_format: file is ascii.\n");
  }
  
/*Checking for statusdict conditions*/
  
  if((is_duplex(param.printer))==TRUE){
    _prints_set_duplex_mode=TRUE;
    if (!(is_noduplex(param.printer))) _prints_format=FALSE;
  }
  if ((pt_format=strstr(format,"-K"))!=NULL){
      stdict_opt_len=4;
      if(*(pt_format+2)=='1'){
        _prints_force_no_duplex=TRUE;
        if (!(is_noduplex(param.printer))) _prints_format=FALSE;
      }	
      else {
        _prints_set_duplex_mode=TRUE;
        if (!(is_noduplex(param.printer))) _prints_format=FALSE;
      }
  } 
  if((is_noduplex(param.printer))==TRUE){
     _prints_set_duplex_mode = FALSE;
     _prints_force_no_duplex = FALSE;
     if(_prints_debug)
       printf("Exec_format: disabling duplex (queue option)\n");
  }
  if((pt_format=strstr(format,"-e"))!=NULL){
     stdict_opt_len+=4;
     _prints_set_tray=TRUE;
     _prints_format=FALSE;
     tray_num = 0;
     if(*(pt_format+2)=='M') _prints_manual_feed=TRUE;
     else if(*(pt_format+2)=='T') _prints_transp_tray=TRUE;
     else if(*(pt_format+2)=='P') 
      {
       _prints_media_pos=TRUE;
       tray_num = ((int)(*(pt_format+3)))-48;
       if (tray_num < 0 || tray_num > 9) tray_num = 0;
      }
     else tray_num=((int)(*(pt_format+2)))-48;
     if(_prints_debug)
       printf("Exec_format: using tray number %c -> %d\n",*(pt_format+2),tray_num);
  }     
  if((is_notray(param.printer))==TRUE){
     _prints_set_tray=FALSE;
     _prints_manual_feed=FALSE;
     _prints_transp_tray=FALSE;
     _prints_media_pos=FALSE;
     if(_prints_debug)
       printf("Exec_format: disabling tray choice (queue option)\n");
  }
  if((is_nolevel1(param.printer))==TRUE){
     _prints_set_nolevel1 = TRUE;
     if(_prints_debug)
       printf("Exec_format: replacing statusdict (L1) with setpagedevice (queue option)\n");
     if ((_prints_set_media_type = media_type(param.printer))!=NULL)
       if(_prints_debug)
         printf("Exec_format: Inserting MediaType %s directive\n",_prints_set_media_type);
  }
  if((strstr(format,"-toff"))!=NULL){
      stdict_opt_len+=6;
      _prints_set_tumble_mode=FALSE;
      if(_prints_debug)
        printf("Exec_format: tumble mode is off\n");
  }	
  if((strstr(format,"-ton"))!=NULL){
      stdict_opt_len+=5;
      _prints_set_tumble_mode=TRUE;
      _prints_format=FALSE;
      if(_prints_debug)
        printf("Exec_format: tumble mode is on\n");
  }

  if((_prints_format==TRUE))
   {
    if(_prints_debug!=0)
       printf("Exec_format: skipping execution of formatting....\n");
    fflush(stdout);   
    free(format);
    free(spool);
    return(0);
   } 

  
/*Executing mpage or psnup... */  

  pt_format=format+stdict_opt_len;
  if(_prints_debug)
    printf("Exec_format: format without statusdict options is: <%s>\n",pt_format);
  
  if(_prints_debug)
    printf("Exec_format: pt_format length is: <%d>\n",strlen(pt_format));

/*No need to run psnup or mpage if only statusdict options are present */
 
  if((strlen(pt_format) < 2)&&(_prints_postscript==TRUE)){
    if(_prints_debug)
      printf("Exec_format: skipping execution of psnup or mpage...\n");
  }else if ((strlen(pt_format) >= 2) || (_prints_postscript==FALSE)){ 
    line_scan = pt_format;
    format_length = strlen(pt_format);
    if(_prints_postscript==TRUE)
     {
      arguments[0]=PRINT_PROGRAM_POST;
      n_args=1;
     }
    else
     {
      arguments[0]=PRINT_PROGRAM_ASCII;

      header_string[0]='\000';
      header_concat[1]='\000';
      for (i=0;i<strlen(param.file);i++)
       {
        header_concat[0]=param.file[i];
        if (param.file[i]=='$') strcat(header_string,"\\");
        if (strlen(header_string)<(sizeof(header_string)-6))
          strcat(header_string,header_concat);
        else break;
       }
      strcat(header_string," (");
      strncat(header_string,param.user,
        sizeof(header_string)-strlen(header_string)-2);
      strcat(header_string,")");
      arguments[1]="-h";
      arguments[2]=header_string;
      n_args=3;  
     }
    while((sscanf(line_scan,"%s",argument)==1)&&n_args<EXEC_FORMAT_MAX_ARGS){
      arguments[n_args] = strstr(line_scan,argument);
      line_scan = arguments[n_args] + strlen(argument);
      *line_scan = '\000';
      if (line_scan < (pt_format+format_length)) line_scan++;
      n_args++;    
     }
       
     if(_prints_debug !=0)
       printf("_prints_format: <%i>\n",
           _prints_format); 	     

     arguments[n_args++] = buf;
     arguments[n_args++] = NULL;

    if (_prints_debug!=0)
      for (i=0;i<n_args;i++) printf("Argument #%02d: <%s>\n",i,arguments[i]);
      
    for (i=n_args;i<EXEC_FORMAT_MAX_ARGS;i++) arguments[i]=NULL;
   
    spool[SPOOL_SIZE-6] = '\000';
    strncpy(spool,buf,SPOOL_SIZE-6);
    strncat(spool,".spl",4);

    cmdout=fopen(spool,"w"); 
    if (cmdout == NULL)
     {
      fprintf(stderr,"Cannot create %s.\n",spool);
      err_sys("exec_format: exiting");     
     }
    if(_prints_debug!=0)
      printf("exec_format: Beginning to fork...\n");
    if ((child_process = fork())<0)
      err_sys("exec_format:Cannot fork");
    if (child_process == 0){          
      if ((dup2(fileno(cmdout),fileno(stdout))) != fileno(stdout))
        err_sys("exec_format: dup failed");     
      if(_prints_postscript==TRUE) 
        exit(execv(PRINT_PROGRAM_POST,arguments));
      else 	 
        exit(execv(PRINT_PROGRAM_ASCII,arguments));    
    } 

    _prints_subprocess_to_cleanup = child_process;

    /* We are in the copy/format child, so we can wait ... */
    if ((retpid=waitpid(child_process, &status, 0)) == -1)
     {
      fprintf(stderr,"exec_format: Error waiting for PID %d: %s \n",
                                  child_process, strerror(errno));
     }

    _prints_subprocess_to_cleanup = 0;

    fclose(cmdout);
    rename(spool,buf);
  }
  free(format);
  
/* We need to execute psselect BEFORE inserting the statusdict
   options, if we want it to be effective */
 
  if (_prints_page_range!=NULL){
     if(_prints_debug !=0)
       printf("print_file:Executing psselect...\n");
     exec_psselect(buf,_prints_page_range);
     free(_prints_page_range);
     _prints_page_range = NULL;
   }

/*Inserting statusdict directives in final postscript file*/

  if((_prints_set_duplex_mode)||(_prints_set_tray)||(_prints_force_no_duplex)) {    
                            /*insert statusdict duplex and tumble mode*/
       if(_prints_debug!=0)
         printf("Exec_format: inserting statusdict directives....\n");
       
       ps_directive_insert_marker[0]=PRINTS_PAGE_DIRECTIVE;
       ps_directive_insert_marker[2]=NULL;

       spool[SPOOL_SIZE-6] = '\000';
       strncpy(spool,buf,SPOOL_SIZE-6);
       strncat(spool,".TMP",4);

       for (n_ps_insert_retries=0;
            !(_prints_set_opt_done)&&n_ps_insert_retries<3;
            n_ps_insert_retries++)
        {
     	    if((spool_file=fopen(buf,"r"))==NULL)
     	       err_sys("exec_format:cannot open input file");
     	    if((cmdout=fopen(spool,"w"))==NULL)
     	       err_sys("exec_format:cannot open output file");
     	    old='\000';
     	    cur='\000';
     	    line_index=0;
     	    while(!feof(spool_file)){
     	      cur=fgetc(spool_file);
     	      if(cur== EOF) break;
     	      if (line_index<(sizeof(line)-1))
     	       {
     		line[line_index++]= cur;
     		line[line_index]='\000';
     	       }
     	      if(old=='\n')
     		{
     /* Workaround for the horrible fig2dev mix of format commands in the
     	PostScript HEADER */
     		 if(strncmp(line,PRINTS_CREATOR_DIRECTIVE,strlen(PRINTS_CREATOR_DIRECTIVE))==0)
                  {
     /* Special recipes for standard violators. */
     		   if (strstr(line,"fig2dev"))
		     ps_directive_insert_marker[0] = PRINTS_END_DIRECTIVE;
     		   if (strstr(line,"HIGZ") || strstr(line,"TRIUMF"))
		     ps_directive_insert_marker[0] = PRINTS_ENDCOMMENTS_DIRECTIVE;
                  }

     		 got_insert_match = FALSE;
                 insert_before_marker = FALSE;
     		 for (cck = 0, cur_check = ps_directive_insert_marker[cck];
     		      cur_check!=NULL;
     		      cck++, cur_check = ps_directive_insert_marker[cck])
     		   if(strncmp(line,cur_check,strlen(cur_check))==0)
     		    {
     		     got_insert_match = TRUE;
                     if ((strcmp(cur_check,PRINTS_PAGE_DIRECTIVE) == 0) ||
                         (strcmp(cur_check,PRINTS_BOGUS_PAGE_DIRECTIVE) == 0))
                      {
                       insert_before_marker = TRUE;
                      }
     		     break;
     		    }
 
     		 if(got_insert_match)
     		   if(_prints_set_opt_done==FALSE)
     		    {
                     if (insert_before_marker)
                      {
                       if (strlen(line)>1)
                         fseek(cmdout, -(strlen(line)-1), 1);
                       fprintf(cmdout,"\n");
                      }

     		     if(_prints_debug!=0)
     		       printf("Exec_format: false condition met...\n");
     		     fflush(stdout);
     		     if((_prints_set_duplex_mode)||(_prints_force_no_duplex)){
     		       if(_prints_debug!=0)
     			  printf("Exec_format: setting duplex mode....\n");
     		       fprintf(cmdout, "%%%%SetDuplexMode\n");
     		       if (_prints_set_duplex_mode)
                        {
                         if (!_prints_set_nolevel1)
     			   fprintf(cmdout,"statusdict begin true setduplexmode end\n");
                         else
     			   fprintf(cmdout,"<</Duplex true>>setpagedevice\n");
                        }
     		       if (_prints_force_no_duplex)
                        {
                         if (!_prints_set_nolevel1)
     			   fprintf(cmdout,"statusdict begin false setduplexmode end\n");
                         else
     			   fprintf(cmdout,"<</Duplex false>>setpagedevice\n");
                        }
     		       fprintf(cmdout, "%%%%EndSetDuplexMode\n");
     		       fprintf(cmdout, "%%%%SetTumbleMode\n");
     		       if(_prints_set_tumble_mode==TRUE){
     			  if(_prints_debug!=0)
     			     printf("Exec_format: setting true tumble mode....\n");
                          if (!_prints_set_nolevel1)
     			    fprintf(cmdout, "statusdict begin true settumble end\n");
                          else
     			    fprintf(cmdout, "<</Tumble true>>setpagedevice\n");
     		       }else{
     			  if(_prints_debug!=0)
     			     printf("Exec_format: setting false tumble mode....\n");
                          if (!_prints_set_nolevel1)
     			    fprintf(cmdout, "statusdict begin false settumble end\n");
                          else
     			    fprintf(cmdout, "<</Tumble false>>setpagedevice\n");
     		       }
     		       fprintf(cmdout, "%%%%EndSetTumbleMode\n");
		       if (_prints_set_nolevel1 && _prints_set_media_type != NULL){
     		         fprintf(cmdout, "%%%%SetMediaType\n");
     		         fprintf(cmdout, "<</MediaType %s >>setpagedevice\n", _prints_set_media_type);
     		         fprintf(cmdout, "%%%%EndSetMediaType\n");
		       }
     		     }
     		     if(_prints_set_tray==TRUE){
     		      if(_prints_manual_feed==TRUE){
     			if(_prints_debug!=0)
     			  printf("Exec_format: setting manual feed....\n");
     			fprintf(cmdout, "%%%%BeginFeature:*ManualFeed\n");
     			fprintf(cmdout, "<</ManualFeed true>>setpagedevice\n");
     			fprintf(cmdout, "%%%%EndFeature\n");
     		      } else if (_prints_transp_tray==TRUE) {
     			if(_prints_debug!=0)
     			  printf("Exec_format: setting transp tray....\n");
     			fprintf(cmdout, "%%%%BeginFeature:*MediaType\n");
     			fprintf(cmdout, "<</MediaType (Transparency)>>setpagedevice\n");
     			fprintf(cmdout, "%%%%EndFeature\n");
     		      } else if (_prints_media_pos==TRUE) {
     			if(_prints_debug!=0)
     			  printf("Exec_format: setting transp tray....\n");
     			fprintf(cmdout, "%%%%BeginFeature:*MediaPosition\n");
     			fprintf(cmdout, "<<\n");
                        fprintf(cmdout, "/MediaPosition %d\n",tray_num);
                        fprintf(cmdout, "/MediaClass null\n");
                        fprintf(cmdout, "/TraySwitch false\n");
                        fprintf(cmdout, "/ManualFeed false\n");
                        fprintf(cmdout,">>setpagedevice\n");
     			fprintf(cmdout, "%%%%EndFeature\n");
     		      } else {
     			 if(_prints_debug!=0)
     			   printf("Exec_format: setting tray number....\n");
     			 fprintf(cmdout, "%%%%SetTray\n");
                         if (!_prints_set_nolevel1)
                          {
     			   fprintf(cmdout, "statusdict /setpapertray known \n");
     			   fprintf(cmdout,"{statusdict begin %d setpapertray end } if\n",
     										  tray_num);
                          }
                         else
                          {
     			   fprintf(cmdout, "<<\n");
                           fprintf(cmdout, "/MediaPosition %d\n",tray_num);
                           fprintf(cmdout, "/MediaClass null\n");
                           fprintf(cmdout, "/TraySwitch false\n");
                           fprintf(cmdout, "/ManualFeed false\n");
                           fprintf(cmdout,">>setpagedevice\n");
                          }
     			 fprintf(cmdout, "%%%%SetTray\n");
     		      }
     		     }
     		    _prints_set_opt_done=TRUE;
                    if (insert_before_marker)
                     {
                      fprintf(cmdout, "%s", line);
                      if (strlen(line)>0)
                        fseek(cmdout, -1, 1);
     		      if(_prints_debug!=0)
                        fprintf(stderr,"Added file <%s> back into file %s.\n",
                                line,buf);
                     }

     		   }
     		  line_index = 0;
     		  line[line_index++]= cur;
     		  line[line_index]='\000';
     		 }
     	       fputc(cur,cmdout);
     	       old=cur;
     	    }
     	    fclose(spool_file);
     	    fclose(cmdout);
	    /* Any luck yet? */
	    if (!(_prints_set_opt_done))
	     {
	      ps_directive_insert_marker[2]=PRINTS_ENDPROLOG_DIRECTIVE;
	      if (n_ps_insert_retries >= 1)
	        ps_directive_insert_marker[2]=PRINTS_BOF_DIRECTIVE;
	     }
        } /* End of insert retrial loop */
       rename(spool,buf);
       if (_prints_debug)
        {
	 if (_prints_set_opt_done)
	   fprintf(stderr,"Inserted PS directives on attempt #%d into file %s.\n",
	           n_ps_insert_retries, buf);
         else
	   fprintf(stderr,"Warning: PS directives NOT inserted into file %s.\n",
	           buf);
	}
  }  
  free(spool);
  if(_prints_debug)
    printf("Exec_format:returning....\n");
  return(0);
}
 
/*
 *  int _prints_client_to_mpage(char* old_format,char *format);
 */
 
int _prints_client_to_mpage(char* old_format,char* format)

{
 char *line_scan;
 char *arguments[EXEC_FORMAT_MAX_ARGS+2];
 char *argument;
 int _prints_boxes=FALSE;
 int _prints_border=FALSE;
 int i,n;
 int format_length,n_args;
 int _prints_set_sides=FALSE;
 char* _prints_sides=NULL;
 char *arg_ptr,arg_str[16];
 int font_size,lines_per_page;
 char *mpage_paper_size = "A4";
 
 if (_prints_page_range != NULL)
  {
   free(_prints_page_range);
   _prints_page_range = NULL;
  }

#define MPAGE_DEFAULT_OPT  " -a"

 argument=(char*)malloc(MAXENTRY);
 *format='\000';
 
 if(_prints_debug !=0)
   printf("_prints_client_to_mpage: format is <%s>\n",old_format);
 
 if ((strcmp(old_format,""))==0){
    free(argument);
    strcat(format, MPAGE_DEFAULT_OPT); /*some default parameters*/
    strcat(format, " -b");
    strcat(format, mpage_paper_size);
    strcat(format," -1 -o"); /*One box only and no border by def.*/
    return 1; 
 }

 if((strstr(old_format,"-@"))!=NULL){
    /*Disabling format options - This flag is only used by the
      LPR interface program (printif), but not by the regular
      client (print)*/
    free(argument);
    return 0; /*no format requested from client*/
 }

 line_scan = old_format;
 format_length = strlen(old_format);
 n_args=0;  
 while((sscanf(line_scan,"%s",argument)==1)&&n_args<EXEC_FORMAT_MAX_ARGS){
    arguments[n_args] = strstr(line_scan,argument);
    n=strlen(argument);
    line_scan = arguments[n_args] +n;
    *line_scan = '\000'; 
    if (line_scan < (old_format+format_length)) line_scan++;
    n_args++;    
 }
   
 /* printf("n_args: %i\n",n_args);*/
 
/*First of all let's put the statusdict options at the beginning...*/
 
 for(i=0;i<n_args;i++){
   if(_prints_debug)
      printf("arguments[%i]=%s\n",i,arguments[i]);  
   if (((strstr(arguments[i],"-K"))!=NULL)||((strstr(arguments[i],"-e"))!=NULL)
        ||((strstr(arguments[i],"-t"))!=NULL)){
     strcat(format," ");
     if (strstr(arguments[i],"-K")!=NULL && strstr(format,"-K")!=NULL) continue;
     if (strstr(arguments[i],"-e")!=NULL && strstr(format,"-e")!=NULL) continue;
     if (strstr(arguments[i],"-t")!=NULL && strstr(format,"-t")!=NULL) continue;
     strncat(format,arguments[i],
        strlen(old_format)-strlen(format)+FORMAT_OFFSET-1); 
   }
 }   
 
 for(i=0;i<n_args;i++){
   if(_prints_debug)
     printf("arguments[%i]=%s\n",i,arguments[i]);
    
   if ((strstr(arguments[i],"-p"))!=NULL){
     if (strstr(format,"-p")!=NULL) continue;
     strcat(format," ");
     strcat(format,arguments[i]);
     continue; 
   } 
   if ((arg_ptr=strstr(arguments[i],"-z"))!=NULL){
     if (strstr(format,"-L")!=NULL) continue;
     font_size = 25;
     sscanf(arg_ptr,"-z%d",&font_size);
     strcat(format," -L");
     lines_per_page = 625/font_size;
     sprintf(arg_str,"%6d\000",lines_per_page);
     strcat(format,arg_str);
     continue;
   } 
   if ((arg_ptr=strstr(arguments[i],"-u"))!=NULL){
     if (strstr(format,"-L")!=NULL) continue;
     lines_per_page = 25;
     sscanf(arg_ptr,"-u%d",&lines_per_page);
     strcat(format," -L");
     sprintf(arg_str,"%6d\000",lines_per_page);
     strcat(format,arg_str);
     continue;
   }
   if ((strstr(arguments[i],"-f"))!=NULL){
     if (strstr(format,"-t")!=NULL) continue;
     strcat(format," -t");
     strcat(format,(arguments[i])+2);
     continue; 
   } 
   if ((strstr(arguments[i],"-B"))!=NULL) {
      if((strstr(arguments[i],"y"))!=NULL) 
       _prints_border=TRUE;	
      continue;
   }    
   if ((strstr(arguments[i],"-N"))!=NULL){
        _prints_boxes=TRUE;
	strcat(format," -");
	strcat(format,(arguments[i])+2);
	/*printf("arguments[%i]+2=%s\n",i,arguments[i]+2);*/
	continue;
   }
   if ((strstr(arguments[i],"-W"))!=NULL) {
      if (strstr(format,"-W")!=NULL) continue;
      strcat(format," ");
      strcat(format,arguments[i]);
      continue;
   }    
   if ((strstr(arguments[i],"-l"))!=NULL){
      if (strstr(format,"-l")!=NULL) continue;
      strcat(format," ");
      strcat(format,arguments[i]);
      continue;
   }  
   if ((strstr(arguments[i],"-h"))!=NULL){
      if (strstr(format,"-h")!=NULL) continue;
      strcat(format," ");
      strcat(format,arguments[i]); 
      continue;
   } 
   if ((strstr(arguments[i],"-P"))!=NULL){
      mpage_paper_size = arguments[i]+2; 
      continue;
   } 
   if ((strstr(arguments[i],"-y"))!=NULL){
      if (_prints_page_range == NULL)
        {
         _prints_page_range = (char *) malloc(strlen(arguments[i]));
         sscanf (arguments[i],"-y%s",_prints_page_range);
         continue;
        }
   }
   /*
    printf("_prints_client_to_mpage: Unrecognized format option: <%s>\n",
                     arguments[i]);  
   */
 }
 
 
 strcat(format, MPAGE_DEFAULT_OPT); /*some default parameters*/
 strcat(format, " -b");
 strcat(format, mpage_paper_size);
    
 if(_prints_boxes == FALSE)
   strcat(format," -1");
 if(_prints_border == FALSE)
   strcat(format," -o");
 
 if (_prints_debug != 0)
   printf(" _prints_client_to_mpage: New format is <%s>\n",format);

 fflush(stdout);
 free(argument);
 return 1;
}



/*
 *  int _prints_client_to_psnup(,char* old_format,char *format);
 */
 
int _prints_client_to_psnup(char* old_format,char* format)
{
 char* line_scan;
 char *arguments[EXEC_FORMAT_MAX_ARGS+2];
 char *argument;
 int _prints_paper=FALSE;
 int _prints_psnup_opt_found=FALSE;
 int i,n;
 int format_length,n_args;
 
 if (_prints_page_range != NULL) 
  {
   free(_prints_page_range);
   _prints_page_range = NULL;
  }

 argument=(char*)malloc(MAXENTRY);

 *format='\000';
 
 if(_prints_debug !=0)
   printf("_prints_client_to_psnup: format is <%s>\n",old_format);
 
 
 if ((strcmp(old_format,""))==0){
    free(argument);
    return 0; /*no format requested from client for postscript file*/
   }
   
 if((strstr(old_format,"-x"))!=NULL){ /*Disabling format options*/  
    free(argument);
    return 0; /*no format requested from client for postscript file*/
   }
   
 if((strstr(old_format,"-@"))!=NULL){ 
    /*Disabling format options - This flag is only used by the
      LPR interface program (printif), but not by the regular
      client (print)*/
    free(argument);
    return 0; /*no format requested from client*/
 }
   
 line_scan = old_format;
 format_length = strlen(old_format);
 n_args=0;  
 while((sscanf(line_scan,"%s",argument)==1)&&n_args<EXEC_FORMAT_MAX_ARGS){
    arguments[n_args] = strstr(line_scan,argument);
    n=strlen(argument);
    line_scan = arguments[n_args] +n;
    *line_scan = '\000'; 
    if (line_scan < (old_format+format_length)) line_scan++;
    n_args++;    
 }
 
 /*First of all let's put the statusdict options at the beginning...*/
 
 for(i=0;i<n_args;i++){
   if(_prints_debug)
      printf("arguments[%i]=%s\n",i,arguments[i]);  
   if (((strstr(arguments[i],"-K"))!=NULL)||((strstr(arguments[i],"-e"))!=NULL)
        ||((strstr(arguments[i],"-t"))!=NULL)){
     if (strstr(arguments[i],"-K")!=NULL && strstr(format,"-K")!=NULL) continue;
     if (strstr(arguments[i],"-e")!=NULL && strstr(format,"-e")!=NULL) continue;
     if (strstr(arguments[i],"-t")!=NULL && strstr(format,"-t")!=NULL) continue;

     strcat(format," ");
     strncat(format,arguments[i],
        strlen(old_format)-strlen(format)+FORMAT_OFFSET-1); 
    }else _prints_psnup_opt_found=TRUE; /*Found a psnup format opt*/
 }   
 
if (_prints_psnup_opt_found==TRUE){
 for(i=0;i<n_args;i++){
   if(_prints_debug)
     printf("arguments[%i]=%s\n",i,arguments[i]);   
   if ((strstr(arguments[i],"-P"))!=NULL){
      if (strstr(format,"-p")!=NULL) continue;
      strcat(format," -p");
      strcat(format,arguments[i]+2);
      _prints_paper=TRUE;
      continue;
   } 
   if ((strstr(arguments[i],"-B"))!=NULL){
      if (strstr(format,"-d")!=NULL) continue;
      if((strstr(arguments[i],"y"))!=NULL) 
          strcat(format," -d");
      continue;
   }   	  
   if ((strstr(arguments[i],"-b"))!=NULL) { /*Should always follow the -d flag!!!*/
      if((strstr(format+strlen(format)-2,"-d"))!=NULL) {
       strcat(format," ");
       strcat(format,arguments[i]+2);
      }
      continue;
   }    
   if ((strstr(arguments[i],"-c"))!=NULL) {
      if (strstr(format,"-c")!=NULL) continue;
      strcat(format," -c");
      continue;
   }   
   if ((strstr(arguments[i],"-d"))!=NULL){
      if (strstr(format,"-s")!=NULL) continue;
      strcat(format," -s");
      strcat(format,arguments[i]+2);
      continue;
   } 
   if ((strstr(arguments[i],"-m"))!=NULL){
      if (strstr(format,"-m")!=NULL) continue;
      strcat(format," -m");
      strcat(format,arguments[i]+2);
      continue;
   } 
   if ((strstr(arguments[i],"-N"))!=NULL){
        if (strstr(format,"-n")!=NULL) continue;
	strcat(format," -n ");
	strcat(format,(arguments[i])+2);
        continue;
   }
   /*
     if ((strstr(arguments[i],"-W"))!=NULL) {
        if (strstr(format,"-w")!=NULL) continue;
        strcat(format," -w");
        strcat(format,arguments[i]+2);
        continue;
     }
   */    
   if ((strstr(arguments[i],"-l"))!=NULL){
      if (strstr(format,"-l")!=NULL) continue;
      strcat(format," -l");
      continue;
   }  
   if ((strstr(arguments[i],"-y"))!=NULL){
      if (_prints_page_range == NULL)
        {
         _prints_page_range = (char *) malloc(strlen(arguments[i]));
         sscanf (arguments[i],"-y%s",_prints_page_range);
         continue;
        }
   }

   /*
     printf("_prints_client_to_psnup: Unrecognized format option: <%s>\n",
                     arguments[i]);  
   */
 }
 
 
 if(_prints_paper==FALSE)
   strcat(format," -Pa4");
 strcat(format," -q");

}
 if(_prints_psnup_opt_found==FALSE){
  if(_prints_debug)
    printf("_prints_client_to_psnup:no psnup option found\n");
 }else{
   if(_prints_debug)
     printf("_prints_client_to_psnup:Found at least one psnup option\n");
 }
 if(_prints_debug)   
   printf(" _prints_client_to_psnup: New format is <%s>\n",format);
 
 fflush(stdout);
 free(argument);
 return 1;
}

int initialize_host_access(allowed_host **head)
 {
  FILE *fin;
  char line[256];
  int column;
  char c;
  int n_tot = 0;
  allowed_host *current_host, *next_host, *prev_host;

  if (*head != NULL)
   {
    /* If this is not the initial call but a SIGHUP call, let's
       first of all get rid of the old hosts */
    for (current_host = *head; current_host != NULL; 
         current_host = next_host)
     {
      next_host = current_host->next;
      free(current_host); 
     }
   }

  *head = NULL;

  /* Open config file */
  if ((fin = fopen(SPOOL_ACCESS_CONFIG,"r"))==NULL)
   {
    if (errno != ENOENT)
     {
      fprintf(stderr,"Warning: cannot access control file %s.\n",
              SPOOL_ACCESS_CONFIG);
      perror("");
     }
    fprintf(stderr, "printd: Running with no global access control.\n");
    return(-1);
   }
  
  prev_host = NULL;
  column = 0;
  c = '\000';

  while (!feof(fin) && c!=EOF)
   {
    c=fgetc(fin);
    if (c=='\r') continue;
    if (c!='\n' && c!=EOF)
     {
      if (column<(sizeof(line)-1)) line[column++]=c;
      continue;
     } 
    if (line[0] == '#')
     {
      column = 0;
      continue;
     }
    if (column==0) continue;
    line[column]='\000';
    next_host = (allowed_host *)malloc(sizeof(allowed_host));
    if (next_host == NULL)
     {
      fprintf(stderr,"Warning: access restriction for host <%s>\n", line);
      fprintf(stderr,"         could not be loaded. Cannot malloc.\n");
      perror("");
      column = 0;
      continue;
     }
    n_tot++;
    if (prev_host != NULL) prev_host->next = next_host;           
    strncpy(next_host->name, line, sizeof(next_host->name));
    next_host->next = NULL;
    prev_host = next_host;
    if (*head == NULL) 
     {
      *head = next_host;
     }
    column = 0;
   }
  
  if (_prints_debug != 0)
   printf("Loaded %d access record(s).\n",n_tot);
  fclose(fin);

  return(0);
 }

void printd_flag_queue_reload()
 {
  _prints_need_to_reload_queues = TRUE;
 }

void printd_toggle_debug()
 {
  time_t s_time;

  s_time = time(0);

  if (_prints_debug != 0)
   {
    _prints_debug = 0;
    printf("End of debug output on %s",ctime(&s_time));
   }
  else
   {
    _prints_debug = 1;
    printf("Starting debug output on %s",ctime(&s_time));
   }
  return;
 }

void printd_exit_cleanup()
 {
  if (_prints_subprocess_to_cleanup != 0)
   {
    kill(_prints_subprocess_to_cleanup, SIGTERM);
    _prints_subprocess_to_cleanup = 0;
   }
  if (getpid()==_prints_parent_pid) spooler_exit_cleanup();
 }

void printd_unlink_temp_file()
 {
  if (_prints_temp_file_to_cleanup!=NULL)
   unlink(_prints_temp_file_to_cleanup);
 }
