 /*
  File :     spooler_queue.c


  Author :   Francesco Prelz ($Author: root $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
   9-Oct-1997 Original release
  10-Feb-1998 Bug fix: Wait for killed processes in order to avoid zombies
  10-Feb-1998 Retrieve output status from file spool child process and
              act accordingly.
  12-Mar-1998 Terminate and wait for file -copy- children as well when
              removing files. Added spooler_queue_reload() to update
              queues on the run. Added spooler_exit_cleanup() to remove
              spool files on crashes.
  20-Mar-1998 Print accounting output for successful jobs.
  31-Mar-1998 Bug fix: need to wait for copy children when a job
              id refused for various reasons.
  24-Jul-1998 Added case for non-standard Olivetti printer LPD
              server which returns "Printer OK" string when
              printer is available.
  24-Jul-1998 Bug fix: keep old max_id queue value when queue is reloaded
              on SIGHUP.
  31-Jul-1998 Clean up secondary if it has been stuck with pending
              jobs for STUCK_SECONDARY_WAIT minutes. 
  10-Sep-1998 Added config file options for disallowing certain format
              options for printers which are not capable of certain
              operations (no_duplex, no_tray, no_size).
  16-Sep-1998 Added lpc_wrap (queue control) and modified accordingly
              lpq_wrap and lpr_wrap and spool loop in order to preserve 
              locked queues.
   5-Nov-1998 Added alias name configuration and handling, with
              dealias_queue_name and get_alias_switches calls.
   6-Sep-1999 Added kludge to wait a certain number of seconds before spooling
              a file to the printer. Works around the behaviour of certain HP
              LPD servers that are unable to handle too many connections 
              per unit time.
  13-Jun-2002 Replaced communication streams with poor man's dynamic strings.
  23-Jan-2003 Added config file options for extra arguments to be passed
              to gs. Added spooler_queue_free (shame on me!).
  12-Jun-2006 Added SPOOLER_QUEUE_NO_LEVEL1 handling.
   9-Jan-2006 Added 'queue empty' message, and config option for
              'empty' message.
  27-Nov-2007 Added optional 'media_type' setting for certain PS-Level2
              printers.
  22-Feb-2008 Added sleep_before_close option.
  20-Jul-2010 Added gs_output_filter option.
  02-May-2016 Accept 'PrinterWarning' as an acceptable status for printing 
              (Xerox Phaser 7500)

  Description: Simple library that implements a certain number
               of printer queues (whose characteristics are read from
               an appropriate configuration file).
               A "round-robin" print call (do_print_wrap) is also implemented,
               which forks the actual print spooling according to the
               queue state.
*/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "spooler_queue.h"
#include "lpd_client.h"
#include "straight_client.h"

spooler_queue *head_spooler_queue=NULL;
alias_queue *head_alias_queue=NULL;
aborted_job *head_aborted_job=NULL;
spooler_queue *dp_save_current_queue=NULL;
char _spooler_queue_config[FILENAME_MAX];
time_t spooler_last_action;

#ifdef TEST_LIBRARY

int _prints_debug = 1;
int _prints_primary = 1;
char *_prints_other_server=NULL;


int main(int argc, char *argv[])
 {
  FILE *lpq_out;
  service_request *null_request=NULL;
  service_request request;
  char id[132];
  char c;

  if (initialize_queues("./test_queues")<0)
   {
    fprintf(stderr,"%s: Cannot initialize queues.\n",argv[0]);
    exit(1);
   }
  
  system("cp ./test_queues ./test_scratch");
  strcpy(request.file,"./test_scratch");
  strcpy(request.user,"prelz");
  request.uid=getuid();
  strcpy(request.host,"osfmi");
  strcpy(request.printer,"ps17_lit3");
  request.reserve = PRINTMI_RESERVE_RES;
  lpq_out = lpr_wrap(&request,"./test_scratch",0);
  fscanf (lpq_out, "Job accepted: ID is %s",id);
  printf("Id from lpr_wrap: %s\n",id);
  close(lpq_out);

  lpq_out = lpq_wrap(null_request);  
  while (!feof(lpq_out))
   {
    c = fgetc(lpq_out);
    if (c == EOF) break;
    putchar(c);
   }
  close(lpq_out);

  strcpy(request.entry,id);
  system("cp ./test_queues ./test_scratch");
  lpq_out = lpr_wrap(&request,"./test_scratch",0);
  while (!feof(lpq_out))
   {
    c = fgetc(lpq_out);
    if (c == EOF) break;
    putchar(c);
   }
  close(lpq_out);
  printf("\nEnd of lpr output.\n");

  lpq_out = lpq_wrap(null_request);
  while (!feof(lpq_out))
   {
    c = fgetc(lpq_out);
    if (c == EOF) break;
    putchar(c);
   }
  close(lpq_out);

  lpq_out = lprm_wrap(&request);
  while (!feof(lpq_out))
   {
    c = fgetc(lpq_out);
    if (c == EOF) break;
    putchar(c);
   }
  close(lpq_out);
  printf("\nEnd of lprm output.\n");
  lpq_out = lpq_wrap(null_request);
  while (!feof(lpq_out))
   {
    c = fgetc(lpq_out);
    if (c == EOF) break;
    putchar(c);
   }
  close(lpq_out);
  printf("\nEnd of output.\n");
  exit(0);
 }
 
/* Stubs for test purposes */
int is_alive(char *server)
 {
  return(0);
 }
int enquire(char *server)
 {
  return(0);
 }
 
#endif

void spooler_queue_reload()
 {
  spooler_queue *old_queue_head;
  spooler_queue *old_queue;
  spooler_queue *current_queue;
  alias_queue *al_queue,*next_al_queue=NULL;
  spooler_queue *following_queue;
  spooler_job   *job,*following_job;
  int child_status;

  /* First of all, get rid of alias queues */
  for(al_queue = head_alias_queue;
      al_queue != NULL; al_queue = next_al_queue)
   {
    next_al_queue = al_queue->next_queue;
    free(al_queue); 
   }
  head_alias_queue = NULL;

  /* Save old queues and generate new queues from the new configuration file */
  old_queue_head = head_spooler_queue;
  head_spooler_queue = NULL;
  initialize_queues(_spooler_queue_config);
  dp_save_current_queue = NULL;

  /* Move existing print jobs to the new queues with the same name */
  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    for(old_queue = old_queue_head;
      old_queue != NULL; old_queue = old_queue->next_queue)
     {
      if (strcasecmp(current_queue->name,old_queue->name)== 0)
       {
        if (old_queue->tail_job != NULL)
         {
          current_queue->tail_job = old_queue->tail_job;
          old_queue->tail_job = NULL;
          current_queue->max_id = old_queue->max_id;
          strncpy(current_queue->status_msg, old_queue->status_msg,
                  sizeof(current_queue->status_msg));
          current_queue->is_locked = old_queue->is_locked;
          old_queue->max_id = 0;
          for (job = current_queue->tail_job;
               job != NULL;
               job = job->prev_job)
            job->queue = current_queue;
         }
        break;
       }
     }
   }
  
  /* Now try to get rid nicely of the old queues, and of the
     print jobs on queues that disappeared from the configuration */

  terminate_aborted_jobs();

  following_queue = NULL;
  for(old_queue = old_queue_head;
    old_queue != NULL; old_queue = following_queue)
   {
    following_queue = old_queue->next_queue;
    following_job = NULL;
    if (old_queue->idlemon_child > 0)
     {
      kill(old_queue->idlemon_child, SIGTERM);
      /* We can afford to wait synchronously here */
      waitpid(old_queue->idlemon_child,&child_status,0);
     }

    for (job = old_queue->tail_job;
         job != NULL;
         job = following_job)
     {
      /* We now remove all jobs (and associated child processes)
         for those queues that don't exist anymore */
      following_job = job->prev_job;
      if (job->copy_child > 0)
       {
        kill(job->copy_child, SIGTERM);
        /* We can afford to wait synchronously here */
        waitpid(job->copy_child,&child_status,0);
       }
      if (job->spool_child > 0)
       {
        kill(job->spool_child, SIGTERM);
        waitpid(job->spool_child,&child_status,0);
       }
      if (strlen(job->temp_file)>0) unlink(job->temp_file);
      free(job);
     }
    spooler_queue_free(old_queue);
   }
  return;
 }

void spooler_exit_cleanup()
 {
  spooler_queue *queue,*following_queue;
  spooler_job   *job,*following_job;
  alias_queue *al_queue,*next_al_queue=NULL;

  for(al_queue = head_alias_queue;
      al_queue != NULL; al_queue = next_al_queue)
   {
    next_al_queue = al_queue->next_queue;
    free(al_queue);
   }
  head_alias_queue = NULL;

  terminate_aborted_jobs();

  following_queue = NULL;
  for(queue = head_spooler_queue;
    queue != NULL; queue = following_queue)
   {
    following_queue = queue->next_queue;
    following_job = NULL;
    for (job = queue->tail_job;
         job != NULL;
         job = following_job)
     {
      /* Just clean up all temp files */
      following_job = job->prev_job;
      if (strlen(job->temp_file)>0) unlink(job->temp_file);
      free(job);
     }
    spooler_queue_free(queue);
   }
  return;
 }

char **parse_arguments(char *start, char *end, int *n_arguments)
 {
  char **arg_array = NULL;
  char *current;
  int arg_len;
  char *cur_arg=NULL;
  int cur_arg_len = 0;
  int inside_quotes = FALSE;
  int i;

  *n_arguments = 0;
  
  arg_len = 0;
  
  for (current = start; ; current++)
   {
    if (*current == '\"')
     {
      if (inside_quotes == FALSE) inside_quotes = TRUE;
      else                        inside_quotes = FALSE;
     }
    /* Pop out double quotes */
    else if ((*current != ' '  && *current != '\n' && 
        *current != '\t' && *current != '\r' ) || (inside_quotes == TRUE))
     {
      /* Non-whitespace case */
      if (arg_len >= cur_arg_len)
       {
        cur_arg_len += 20;
        cur_arg = (char *)realloc(cur_arg, cur_arg_len+1); 
        if (cur_arg == NULL)
         {
          fprintf(stderr,"Out of memory.\n");
          /* Get rid of what we already had */
          for (i=0; i<((*n_arguments)-1); i++)
           {
            free(arg_array[i]);
           }
          if (arg_array != NULL) free(arg_array);
          *n_arguments = 0;
          return(NULL);
         }
       }
      cur_arg[arg_len]=*current;
      arg_len++;
      cur_arg[arg_len]='\000';
     }

    if ((inside_quotes == FALSE) && (*current == ' '  || *current == '\n' || 
        *current == '\t' || *current == '\r') || current == end)
     {
      /* Whitespace or end-of-string case */
      if (arg_len > 0)
       {
        /* Add argument to list */
        (*n_arguments)++;
        arg_array = (char **)realloc(arg_array, sizeof(char *)*(*n_arguments));
        if (arg_array == NULL)
         {
          *n_arguments = 0;
          fprintf(stderr,"Out of memory.\n");
          /* We lost the internal pointers, too... */
          free(cur_arg);
          return(NULL);
         }
        arg_array[(*n_arguments)-1] = strdup(cur_arg);
        arg_len = 0;
       }
      if (current == end) break;
     }
   }
  if (cur_arg != NULL) free(cur_arg);
  return(arg_array); 
 }

void spooler_queue_free(spooler_queue *q)
 {
  int i;
  if (q==NULL) return;

  if (q->user_access_allow != NULL) free(q->user_access_allow);
  if (q->host_access_allow != NULL) free(q->host_access_allow);
  if (q->user_access_deny  != NULL) free(q->user_access_deny);
  if (q->host_access_deny  != NULL) free(q->host_access_deny);
  if (q->description       != NULL) free(q->description);
  if (q->file_check_string != NULL) free(q->file_check_string);
  if (q->empty_check_string != NULL) free(q->empty_check_string);
  if (q->media_type != NULL)        free(q->media_type);
  if (q->gs_option         != NULL) free(q->gs_option);
  if (q->gs_extra_arguments != NULL)
   {
    for (i=0; i<(q->gs_extra_arguments_n); i++)
     {
      if (q->gs_extra_arguments[i] != NULL)
        free(q->gs_extra_arguments[i]);
     }
    free(q->gs_extra_arguments);
   }
  if (q->gs_output_filter != NULL) free(q->gs_output_filter);
  
  free(q);
  return;
 }

int initialize_queues(char *file_name)
 {
  char *config_file;
  FILE *config_fd;
  char line[1024];
  char entry[512];
  int line_cursor;
  char *line_scan, *value;
  char *start_extra_gs_args, *end_extra_gs_args;
  char *start_gs_output_filter;
  int n_filter_args;
  int ifr;
  spooler_queue *new_spooler_queue,*current_queue;
  spooler_queue **current_spooler_queue;
  alias_queue *new_alias_queue;
  alias_queue **current_alias_queue;
  char *first_quote, *second_quote;
  int quoted_len;
  int total_queues;
  char c,prev_c;
  char **filter_args;
  
  if (file_name == NULL)
   {
    /* Try to get the config filename from the environment */
    if ((config_file = getenv("PRINTMI_QUEUE_CONFIG"))==NULL)
    fprintf(stderr,"Cannot determine spooler queue configuration.\n");
    fprintf(stderr,"No queues loaded. Try setting PRINTMI_QUEUE_CONFIG.\n");
    return(-1);
   }
  else
   {
    if ((config_file = getenv("PRINTMI_QUEUE_CONFIG"))==NULL)
      config_file = file_name;
   }

  strncpy(_spooler_queue_config,config_file,sizeof(_spooler_queue_config));

  if ((config_fd = fopen(config_file,"r"))==NULL)
   {
    fprintf(stderr,"Cannot open config file %s: %s\n",config_file,
            strerror(errno));
    fprintf(stderr,"No queues loaded.\n");
    return(-1);
   }

  current_spooler_queue = &head_spooler_queue;
  current_alias_queue = &head_alias_queue;
  line_cursor=0;
  total_queues = 0;
  c = 0;
  while (!feof(config_fd))
   {
    prev_c = c;
    c = fgetc(config_fd);
    if ((c=='\n')||(c==EOF))
     {
      if (prev_c == '\\') /* Handle continuation character */
       {
        if (line_cursor>0) line_cursor--;
        continue;
       }

      line[line_cursor]='\000';
      line_cursor = 0;

      if ((line[0]!='#') && (line[0]!='!'))
       {
        if (total_queues>=MAX_TOTAL_QUEUES)
         {
          fprintf(stderr, "Warning: the maximum number of queues (%d)"
                          " was reached. Discarding:\n", MAX_TOTAL_QUEUES);
          fprintf(stderr, "%s\n",line);
          continue;
         }

        /* get local queue name */
        if (sscanf(line,"%s",entry)!=1)
         {
          continue;
         }
        line_scan = strstr(line,entry) + strlen(entry);

        /* Is this a queue alias? */
#define ALIAS_SIGNATURE "alias:"
        if (strncmp(entry,ALIAS_SIGNATURE,strlen(ALIAS_SIGNATURE))== 0)
         {
          /* Create alias */
          /* get real queue name */
          if (sscanf(line_scan,"%s",entry)!=1)
           {
            continue;
           }
	  line_scan = strstr(line_scan,entry) + strlen(entry);
          /* Does this queue exist? */
          for(current_queue = head_spooler_queue; current_queue != NULL;
              current_queue = current_queue->next_queue)
           {
            if (strcmp(current_queue->name,entry)==0) break;
           }
          if (current_queue == NULL) 
           {
            fprintf(stderr,"Warning: real queue %s not found. Skipping alias.\n",
                    entry);
            continue;
           }
          new_alias_queue = (alias_queue *)malloc(sizeof(alias_queue));
          if (new_alias_queue == NULL)
           {
            fprintf(stderr,"Warning: cannot malloc. Skipping alias queue.\n");
            continue;
           }
          new_alias_queue->real_queue = current_queue;
          /* get alias queue name */
          if (sscanf(line_scan,"%s",entry)!=1)
           {
            free(new_alias_queue);
            continue;
           }
          line_scan = strstr(line_scan,entry) + strlen(entry);
          strncpy(new_alias_queue->name,entry,sizeof(new_alias_queue->name));
          while (*line_scan==' ' || *line_scan=='\t') line_scan++;
          strncpy(new_alias_queue->switches,line_scan,sizeof(new_alias_queue->switches));
          new_alias_queue->next_queue = NULL;
          current_queue->options |= SPOOLER_QUEUE_HAS_ALIASES;
          *current_alias_queue = new_alias_queue;
          current_alias_queue = &(new_alias_queue->next_queue);
          continue;
         }

        new_spooler_queue = (spooler_queue *)malloc(sizeof(spooler_queue));
        if (new_spooler_queue == NULL)
         {
          fprintf(stderr,"Warning: cannot malloc. Skipping queue %s.\n",entry);
          continue;
         }
        new_spooler_queue->tail_job = NULL;
        new_spooler_queue->next_queue = NULL;
        new_spooler_queue->busy_until = time(0);
        new_spooler_queue->max_id = 0;
        new_spooler_queue->is_locked = FALSE;
        new_spooler_queue->status_msg[0] = '\000';
        new_spooler_queue->idlemon_child = -1;

        strncpy(new_spooler_queue->name,entry,sizeof(new_spooler_queue->name));
        /* get lpd server name */
        if (sscanf(line_scan,"%s",entry)!=1)
         {
          free(new_spooler_queue);
          continue;
         }
        line_scan = strstr(line_scan,entry) + strlen(entry);
        strncpy(new_spooler_queue->lpd_server, entry,
               sizeof(new_spooler_queue->lpd_server));
        /* get lpd (remote) queue name */
        if (sscanf(line_scan,"%s",entry)!=1)
         {
          free(new_spooler_queue);
          continue;
         }

	line_scan = strstr(line_scan,entry) + strlen(entry);
        strncpy(new_spooler_queue->lpd_queue, entry,
               sizeof(new_spooler_queue->lpd_queue));
        new_spooler_queue->telnet_port = 0;
        if (strncmp(new_spooler_queue->lpd_queue,"tcp_port:",
                    strlen("tcp_port:"))==0)
         {
          sscanf(new_spooler_queue->lpd_queue,"tcp_port:%d",
                 &new_spooler_queue->telnet_port);
         }
	
        /* Initialize "idle queue" monitor */
	new_spooler_queue->idle_monitor = 0;
	if ((value = strstr(line_scan,"idlemon"))!=NULL)
	 {
	  value += strlen("idlemon") + 1;	  
	  sscanf(value,"%d",&new_spooler_queue->idle_monitor);
	  new_spooler_queue->idle_monitor *= 60; /*value is in minutes*/
	 }
	  
        /* Initialize user- or hostname- based access control masks */
        new_spooler_queue->user_access_allow=NULL;
        new_spooler_queue->host_access_allow=NULL;
        new_spooler_queue->user_access_deny=NULL;
        new_spooler_queue->host_access_deny=NULL;

        if ((value = strstr(line_scan,"allow_user"))!=NULL)
         {
          value += strlen("allow_user") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->user_access_allow = 
              (char *)malloc(strlen(entry)+1);
            if (new_spooler_queue->user_access_allow != NULL)
              strcpy(new_spooler_queue->user_access_allow,entry);
           }
         }

        if ((value = strstr(line_scan,"allow_host"))!=NULL)
         {
          value += strlen("allow_host") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->host_access_allow =
              (char *)malloc(strlen(entry)+2);
            if (new_spooler_queue->host_access_allow != NULL)
              sprintf(new_spooler_queue->host_access_allow,"%s,",entry);
           }
         }

        if ((value = strstr(line_scan,"deny_user"))!=NULL)
         {
          value += strlen("deny_user") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->user_access_deny =
              (char *)malloc(strlen(entry)+1);
            if (new_spooler_queue->user_access_deny != NULL)
              strcpy(new_spooler_queue->user_access_deny,entry);
           }
         }

        if ((value = strstr(line_scan,"deny_host"))!=NULL)
         {
          value += strlen("deny_host") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->host_access_deny =
              (char *)malloc(strlen(entry)+2);
            if (new_spooler_queue->host_access_deny != NULL)
              sprintf(new_spooler_queue->host_access_deny,"%s,",entry);
           }
         }

        /* Get printer description, if any */
        new_spooler_queue->description=NULL;
        if ((value = strrchr(line_scan,'|'))!=NULL)
         {
          value++;
          new_spooler_queue->description =
            (char *)malloc(strlen(line_scan)-(int)(value-line_scan)+1);
          if (new_spooler_queue->description != NULL)
            strcpy(new_spooler_queue->description,value);
         }

        /* Initialize ghostscript filter */
        new_spooler_queue->gs_option=NULL;
        new_spooler_queue->gs_extra_arguments=NULL;
        new_spooler_queue->gs_extra_arguments_n=0;
        new_spooler_queue->gs_output_filter=NULL;

        if ((value = strstr(line_scan,"use_gs"))!=NULL)
         {
          value += strlen("use_gs") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->gs_option =
              (char *)malloc(strlen(entry)+1);
            if (new_spooler_queue->gs_option != NULL)
              strcpy(new_spooler_queue->gs_option,entry);
           }
          if ((start_extra_gs_args = strstr(line_scan,"start_extra_gs_args"))!=NULL)
           {
            start_extra_gs_args += strlen("start_extra_gs_args")+1;
            if ((end_extra_gs_args = strstr(line_scan,"end_extra_gs_args"))!=NULL)
             {
              end_extra_gs_args -= 1;
              new_spooler_queue->gs_extra_arguments = parse_arguments(
                  start_extra_gs_args, end_extra_gs_args, 
                  &(new_spooler_queue->gs_extra_arguments_n)); 
             }
           }
          if ((start_gs_output_filter = strstr(line_scan,"gs_output_filter"))!=NULL)
           {
            start_gs_output_filter += strlen("gs_output_filter")+1;
            filter_args = parse_arguments(start_gs_output_filter,
                                          line_scan + strlen(line_scan),
                                          &n_filter_args);
            if (n_filter_args > 0)
             {
              new_spooler_queue->gs_output_filter = filter_args[0];
              for (ifr=1; ifr<n_filter_args; ifr++) free(filter_args[ifr]);
              free(filter_args);
             }
           }
         }

        /* Seconds to sleep before spooling file (kludge for certain HP
           printers) */
        new_spooler_queue->sleep_before_spool=0;

        if ((value = strstr(line_scan,"sleep_before_spool"))!=NULL)
         {
          value += strlen("sleep_before_spool") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->sleep_before_spool =
              atoi(entry);
           }
         }

        /* Seconds to sleep before closing a passthrough TCP
           connection (kludge for Ricoh printers) */
        new_spooler_queue->sleep_before_close=0;

        if ((value = strstr(line_scan,"sleep_before_close"))!=NULL)
         {
          value += strlen("sleep_before_close") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->sleep_before_close =
              atoi(entry);
           }
         }

        /* Get "file" check option, if any */
        new_spooler_queue->file_check_string=NULL;

        if ((value = strstr(line_scan,"check_file"))!=NULL)
         {
          value += strlen("check_file") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->file_check_string =
              (char *)malloc(strlen(entry)+1);
            if (new_spooler_queue->file_check_string != NULL)
              strcpy(new_spooler_queue->file_check_string,entry);
           }
         }

        /* Get "empty" check string, if any */
        new_spooler_queue->empty_check_string=NULL;

        if ((value = strstr(line_scan,"empty_check"))!=NULL)
         {
          value += strlen("empty_check") + 1;
          first_quote = strchr(value,'"');
          if (first_quote != NULL)
           { 
            second_quote = strchr(first_quote+1,'"');
            if (second_quote != NULL)
             {
              quoted_len = (int)(second_quote - first_quote);
              new_spooler_queue->empty_check_string =
                (char *)malloc(quoted_len);
              if (new_spooler_queue->empty_check_string != NULL)
               {
                memcpy(new_spooler_queue->empty_check_string,
                       first_quote+1, quoted_len-1);
                new_spooler_queue->empty_check_string[quoted_len-1] = '\000';
               }
             }
           }
         }

        /* Get media_type option, if any */
        new_spooler_queue->media_type=NULL;

        if ((value = strstr(line_scan,"media_type"))!=NULL)
         {
          value += strlen("media_type") + 1;
          sscanf(value,"%s",entry);
          if (strlen(entry)>0)
           {
            new_spooler_queue->media_type =
              (char *)malloc(strlen(entry)+1);
            if (new_spooler_queue->media_type != NULL)
              strcpy(new_spooler_queue->media_type,entry);
           }
         }

	new_spooler_queue->options = 0;
	
        /* Look for queue options */
	if (strstr(line_scan,"reserve")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_RESV;
	if (strstr(line_scan,"reorder")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_SORT;
	if (strstr(line_scan,"banner")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_BANN;
	if (strstr(line_scan,"duplex")!=NULL)
         {
	  if (strstr(line_scan,"noduplex")!=NULL)
	    new_spooler_queue->options |= SPOOLER_QUEUE_NODU;
          else
	    new_spooler_queue->options |= SPOOLER_QUEUE_DUPL;
         }
	if (strstr(line_scan,"account")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_ACCT;
	if (strstr(line_scan,"nostatus")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_NSTA;
	if (strstr(line_scan,"just_postscript")!=NULL)
	  new_spooler_queue->options |= SPOOLER_QUEUE_JPSF;
        if (strstr(line_scan,"notray")!=NULL)
          new_spooler_queue->options |= SPOOLER_QUEUE_NOTR;
        if (strstr(line_scan,"lpdstatus")!=NULL)
          new_spooler_queue->options |= SPOOLER_QUEUE_LPDS;
        if (strstr(line_scan,"nolevel1")!=NULL)
          new_spooler_queue->options |= SPOOLER_QUEUE_NO_LEVEL1;
	  
        *current_spooler_queue = new_spooler_queue; 
        current_spooler_queue = &(new_spooler_queue->next_queue);
        total_queues++;
       }
      if (c != EOF) continue;
     }
    if (c == EOF) break;
    if (line_cursor<(sizeof(line)-1)) line[line_cursor++] = c;
   }
      
  fclose(config_fd);
  spooler_last_action = time(0);
  return(0);
 }
 
int is_duplex(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      if (((current_queue->options)&SPOOLER_QUEUE_DUPL)!=0) return(TRUE);
      else return(FALSE);
     }
   }
  /* Default to duplex printing for non-existing queues (Well...) */
  return(TRUE);
 }

int is_noduplex(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      if (((current_queue->options)&SPOOLER_QUEUE_NODU)!=0) return(TRUE);
      else return(FALSE);
     }
   }
  /* Default to duplex printing for non-existing queues (Well...) */
  return(FALSE);
 }

int is_notray(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      if (((current_queue->options)&SPOOLER_QUEUE_NOTR)!=0) return(TRUE);
      else return(FALSE);
     }
   }
  /* Default to no trays for non-existing queues (Well...) */
  return(TRUE);
 }

int is_nolevel1(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      if (((current_queue->options)&SPOOLER_QUEUE_NO_LEVEL1)!=0) return(TRUE);
      else return(FALSE);
     }
   }
  /* Default to Level1 for non-existing queues (Well...) */
  return(FALSE);
 }

int wants_banner(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      if (((current_queue->options)&SPOOLER_QUEUE_BANN)!=0) return(TRUE);
      else return(FALSE);
     }
   }
  /* Default to no banner for non-existing queues (Well...) */
  return(FALSE);
 }

char *use_gs(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
      return (current_queue->gs_option);
   }
  /* Default to no gs for non-existing queues (Well again...) */
  return(NULL);
 }

char **extra_gs_arguments(char *queue_name, int *n_arguments)
 {
  spooler_queue *current_queue;

  *n_arguments = 0;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
     {
      *n_arguments = current_queue->gs_extra_arguments_n;
      return (current_queue->gs_extra_arguments);
     }
   }
  /* Default to no extra arguments for non-existing queues. */
  return(NULL);
 }

char *gs_output_filter(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
      return (current_queue->gs_output_filter);
   }
  return(NULL);
 }

char *check_file(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
      return (current_queue->file_check_string);
   }
  return(NULL);
 }

char *media_type(char *queue_name)
 {
  spooler_queue *current_queue;

  for(current_queue=head_spooler_queue; current_queue!=NULL;
      current_queue = current_queue->next_queue)
   {
    if (strcasecmp(current_queue->name,queue_name)==0)
      return (current_queue->media_type);
   }
  return(NULL);
 }

char *dealias_queue_name(char *name)
 {
  alias_queue *al_queue;

  if (strlen(name)<=0) return (name);

  for (al_queue = head_alias_queue; al_queue != NULL;
       al_queue = al_queue->next_queue)
   {
    if (strcasecmp(al_queue->name,name) == 0)
     {
      return (al_queue->real_queue->name);
     }
   }
  return (name);
 }

char *get_alias_switches(char *name)
 {
  alias_queue *al_queue;

  if (strlen(name)<=0) return (NULL);

  for (al_queue = head_alias_queue; al_queue != NULL;
       al_queue = al_queue->next_queue)
   {
    if (strcasecmp(al_queue->name,name) == 0)
     {
      return (al_queue->switches);
     }
   }
  return (NULL);
 }

DSTRING dsprintf(DSTRING ds, char * fmt, ... )
 {
  int alloc_size; 
  int print_size;
  int printed_size;
  va_list ap;
  va_start(ap, fmt);
  if (ds == NULL)
   {
    alloc_size = 2*sizeof(int) + DSPRINTF_INCREASE_SIZE;
    ds= (int *)malloc( alloc_size );
    if (ds==NULL) return (NULL); 
    ds[0]=2*sizeof(int);
    ds[1]=alloc_size; 
    *(DSTRING_PTR(ds)) = '\000';
   }

  while (print_size=ds[1]-ds[0],
         (printed_size = vsnprintf((char *)(ds)+(*ds),print_size,fmt,ap) )
           >= print_size)
   {
    alloc_size = ds[1]+DSPRINTF_INCREASE_SIZE;
    if (alloc_size > DSPRINTF_MAX_SIZE) return (ds);
    ds = realloc(ds, alloc_size);
    if (ds==NULL) return (NULL); 
    ds[1]=alloc_size;
    if (_prints_debug != 0)
      printf("dsprintf: Realloc. Writing at %d. Alloc %d. \n", ds[0],ds[1]);
   }
  ds[0] += printed_size;
  return(ds);
 }

DSTRING lpq_wrap(service_request *svreq)
 {
  DSTRING rtdbuf = NULL;
  spooler_queue *current_queue;
  spooler_job *current_job;
  alias_queue *al_queue;
  char *req_printer;
  int jobs_on_queue;
  int queue_found = FALSE;
  time_t reserve_exp;
  int status;
  int i;

  if (_prints_debug != 0) printf("lpq: requested printer <%s>\n",svreq->printer);
  req_printer = dealias_queue_name(svreq->printer);
  if (_prints_debug != 0 && req_printer != svreq->printer) 
    printf("lpq: real printer <%s>\n",req_printer);

  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    if (svreq!=NULL)
      if (req_printer!=NULL&&req_printer!="")
        if ((strlen(req_printer)>0)&&(strcasecmp(current_queue->name,req_printer)!=0)) continue;

    queue_found = TRUE;        

    if (svreq->id_only)
     {
      if (current_queue->description!=NULL)
        rtdbuf=dsprintf(rtdbuf,"%s|%s\n",current_queue->name,
                                     current_queue->description);
      else rtdbuf=dsprintf(rtdbuf,"%s|\n",current_queue->name);

      /* Does this queue have any alias names? */
      if (((current_queue->options)&SPOOLER_QUEUE_HAS_ALIASES) != 0)
       {
        for (al_queue = head_alias_queue; al_queue != NULL;
             al_queue = al_queue->next_queue)
         {
          if (al_queue->real_queue == current_queue)
           {
            rtdbuf=dsprintf(rtdbuf,"%s|Alias for %s (%s)\n",al_queue->name,
                                   current_queue->name,al_queue->switches);
           }
         }
       }

      continue;
     }

    rtdbuf=dsprintf(rtdbuf,"Queue: %s", current_queue->name);
    if (current_queue->description!=NULL)
      rtdbuf=dsprintf(rtdbuf," (%s)\n", current_queue->description);
    else
      rtdbuf=dsprintf(rtdbuf,"\n");
    if ((strlen(current_queue->status_msg)>0) &&
        ( ((current_queue->options & SPOOLER_QUEUE_LPDS) != 0) ||
          current_queue->is_locked ) )
      rtdbuf=dsprintf(rtdbuf,"%s\n", current_queue->status_msg);

#if 0
    if (_prints_debug != 0)
     {
      rtdbuf=dsprintf(rtdbuf,"Queue options: %d Idle monitor (s): %d\n",
              current_queue->options, current_queue->idle_monitor);
      rtdbuf=dsprintf(rtdbuf,"Allow: <%s> <%s>\n",
              current_queue->user_access_allow, current_queue->host_access_allow);
      rtdbuf=dsprintf(rtdbuf,"Deny:  <%s> <%s>\n",
              current_queue->user_access_deny,  current_queue->host_access_deny);
     }
#endif

    /* Does this queue have any alias names? */
    if (((current_queue->options)&SPOOLER_QUEUE_HAS_ALIASES) != 0)
     {
      rtdbuf=dsprintf(rtdbuf,"Aliases:\n");
      for (al_queue = head_alias_queue; al_queue != NULL;
           al_queue = al_queue->next_queue)
       {
        if (al_queue->real_queue == current_queue)
         {
          rtdbuf=dsprintf(rtdbuf,"  %s (%s)\n",al_queue->name,al_queue->switches);
         }
       }
     }

    current_job = current_queue->tail_job;

    if (current_queue->is_locked)
     {
      rtdbuf=dsprintf(rtdbuf,"This queue is currently not accessible.\n");
     }
    else if (current_job==NULL)
     {
      rtdbuf=dsprintf(rtdbuf,"No print jobs.\n");
     }
    else
     {
      jobs_on_queue = 0;
      rtdbuf=dsprintf(rtdbuf,"Print jobs:\n");
      /* First traverse the list backwards */
      while(current_job->prev_job!=NULL)
        current_job = current_job->prev_job;
	
      /* Check whether the first job in queue is a reservation
         placeholder */
      if (current_job!=NULL)
        if (current_job->is_reservation)
	 {
	  reserve_exp = current_queue->last_action+MAX_RESERVE_SECONDS;
	  rtdbuf=dsprintf(rtdbuf,"Queue is reserved and on hold until %s",
	          ctime(&reserve_exp));
         }
	 
      for(;current_job!=NULL;current_job=current_job->next_job)
       {
        jobs_on_queue++;
	if (current_job->is_reservation)
	 {
	  rtdbuf=dsprintf(rtdbuf, "%2d) %s/%d\t Reservation by %s@%s\n",
	          jobs_on_queue, current_job->queue->name,
	          current_job->queue_no,
                  current_job->user, current_job->host);
          continue;
	 }
	 
        rtdbuf=dsprintf(rtdbuf,"%2d) %s/%-3d %-15s %s@%s Size: %d ",
               jobs_on_queue, current_job->queue->name,
	       current_job->queue_no,
	       current_job->orig_file,
               current_job->user, current_job->host,
               current_job->file_size);
        if (current_job->copies_to_print>1)
	  rtdbuf=dsprintf(rtdbuf,"(%dx) ",current_job->copies_to_print);
	switch(current_job->status)
	 {
	  case waiting:
	   rtdbuf=dsprintf(rtdbuf,"WAIT \n");
	   break;
	  case rescop:
	  case copying:
	   rtdbuf=dsprintf(rtdbuf,"XSFER\n");
	   break;
	  case resrdy:
	  case ready:
	   rtdbuf=dsprintf(rtdbuf,"READY\n");
	   break;
	  case error:
	   rtdbuf=dsprintf(rtdbuf,"ERROR\n");
	   break;
	  case printing:
	   rtdbuf=dsprintf(rtdbuf,"PRINT\n");
	 }
	   
       }      
      rtdbuf=dsprintf(rtdbuf,"\n");
     }
   }
  if (!(queue_found))
   {
    if (strlen(req_printer)==0) rtdbuf=dsprintf(rtdbuf,"No queues configured!\n");
    else rtdbuf=dsprintf(rtdbuf,"Queue %s not found.\n",req_printer);
   }

  return(rtdbuf);
 }

DSTRING lpc_wrap(service_request *svreq)
 {
  DSTRING rtdbuf = NULL;
  char *req_printer;
  spooler_queue *current_queue;
  int queue_found = FALSE;
  int status;
  int i;

  req_printer = dealias_queue_name(svreq->printer);

  if (_prints_debug != 0) printf("lpc: requested printer <%s>\n",req_printer);
  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    if (svreq!=NULL)
      if (svreq->printer!=NULL&&svreq->printer!="")
        if ((strlen(req_printer)>0)&&(strcasecmp(current_queue->name,req_printer)!=0)) continue;

    queue_found = TRUE;

    if (svreq->reserve == PRINTMI_LPC_DISABLE)
     {
      strncpy(current_queue->status_msg,svreq->lpc_message,
              (sizeof(current_queue->status_msg)-1));
      current_queue->status_msg[sizeof(current_queue->status_msg)-1] = '\000';
      if (current_queue->is_locked)
       {
        rtdbuf=dsprintf(rtdbuf,"Queue %s already disabled. Message updated.\n",
                current_queue->name);
       }
      else
       {
        current_queue->is_locked = TRUE;
        rtdbuf=dsprintf(rtdbuf,"Queue %s is now disabled.\n",
                current_queue->name);
       }
     }
    else
     {
      strncpy(current_queue->status_msg,svreq->lpc_message,
              sizeof(current_queue->status_msg));
      if (current_queue->is_locked)
       {
        current_queue->is_locked = FALSE;
        rtdbuf=dsprintf(rtdbuf,"Queue %s is now enabled.\n",
                current_queue->name);
       }
      else
       {
        rtdbuf=dsprintf(rtdbuf,"Queue %s already enabled. Message updated.\n",
                current_queue->name);
       }
     }
   }

  if (!(queue_found))
   {
    if (req_printer == NULL) rtdbuf=dsprintf(rtdbuf,"No queue requested!\n");
    else rtdbuf=dsprintf(rtdbuf,"Queue %s not found.\n",req_printer);
   }
  return(rtdbuf);
 }

DSTRING lprm_wrap(service_request *svreq)
 {
  DSTRING rtdbuf = NULL;
  spooler_queue *current_queue;
  spooler_job *current_job;
  char req_queue[MAX_QUEUE_NAME_LEN+10];
  char *slash_pos;
  int req_id;
  int jobs_on_queue;
  int i,status;

  if (_prints_debug!=0)
    printf("lprm: now trying to remove %s\n",svreq->entry);

  if (sscanf(svreq->entry,"%s",req_queue)!=1)
   {
    rtdbuf=dsprintf(rtdbuf,"lprm: %s is an invalid entry\n",
            svreq->entry);
    return(rtdbuf);
   }
  slash_pos = strchr(req_queue,'/');
  if (slash_pos == NULL)
   {
    rtdbuf=dsprintf(rtdbuf,"lprm: %s is an invalid entry\n",
            svreq->entry);
    return(rtdbuf);
   }
  req_id = atoi(slash_pos+1);
  *slash_pos = '\000';
  if (req_id ==0)
   {
    rtdbuf=dsprintf(rtdbuf,"lprm: %s is an invalid entry\n",
            svreq->entry);
    return(rtdbuf);
   }
   
  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    if (strstr(current_queue->name,req_queue)==NULL) continue;
        
    current_job = current_queue->tail_job;
    
    if (current_job!=NULL)
     {
      for(;current_job!=NULL;current_job=current_job->prev_job)
       {
        if (current_job->queue_no == req_id)
         {
          if (svreq->uid!=0 && current_job->user_uid!=JAVA_CLIENT_UID &&
              ((svreq->uid!=current_job->user_uid) ||
              (strcasecmp(current_job->user,svreq->user)!=0)))
           {
            rtdbuf=dsprintf(rtdbuf,"lprm: Cannot remove %s: not owner (%s/%d)\n",
                    svreq->entry, current_job->user, current_job->user_uid);
            return(rtdbuf);
           }

          do_dequeue_and_remove_job(current_job, svreq->uid, svreq->user);

          rtdbuf=dsprintf(rtdbuf,"lprm: %s successfully dequeued\n",
                  svreq->entry);
          return(rtdbuf);
         }
       }      
     }
   }

  if (_prints_primary != 0)
    if (is_alive(_prints_other_server)) 
      if (enquire(_prints_other_server)>0) 
       {
        rtdbuf=dsprintf(rtdbuf,"lprm: Job not found on this server.\n"
                            "      Try lprm -S %s\n",svreq->entry);
        return(rtdbuf);
       }

  rtdbuf=dsprintf(rtdbuf,"lprm: %s not found.\n",svreq->entry);
  return(rtdbuf);
 }

void reset_all_jobs()
 {
  spooler_queue *current_queue;
  spooler_job *current_job;

  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    current_job = current_queue->tail_job;
   
    if (current_job!=NULL)
     {
      for(;current_job!=NULL;current_job=current_job->prev_job)
        do_dequeue_and_remove_job(current_job, 0, "");
     }
   }
 }

void do_dequeue_and_remove_job(spooler_job *job, int uid, char *user)
 {
  static int remove_child = -1;
  int child_to_terminate = -1;
  int i,status,child_status;
  uid_t eff_uid;
  gid_t eff_gid;

  if (job->copy_child > 0)
    child_to_terminate = job->copy_child;
  if (job->spool_child > 0)
    child_to_terminate = job->spool_child;

  if (child_to_terminate >0)
   {
    if ((status=waitpid(child_to_terminate,&child_status,WNOHANG))>0)
     {
      if (job->status == printing) job->spool_child = -1;
      if (job->status == copying) job->copy_child = -1;
     }
    else if (status == 0)
     {
      for (i=0;i<3;i++)
       {
        kill(child_to_terminate, SIGTERM);
        /* Wait for killed process: try few times waiting for
           processes stuck in kernel mode */
        if (waitpid(child_to_terminate,&child_status,WNOHANG)>0)
          break;
        sleep(1);
        if (waitpid(child_to_terminate,&child_status,WNOHANG)>0)
          break;
       }
     }
   }

  eff_uid = geteuid();
  eff_gid = getegid();

  /* restore 0 UID to be able to open LPD port */
  if (setreuid(-1,0)!=0)
    fprintf(stderr,"Cannot set UID for LPD port access\n");
  if (setregid(-1,0)!=0)
    fprintf(stderr,"Cannot set GID for LPD port access\n");

  if (job->queue->telnet_port==0)
   {
    if (remove_child > 0)
     {
      for (i=0;i<3;i++)
       {
        if (waitpid(remove_child,&child_status,WNOHANG)>0)
          break;
        kill(remove_child, SIGTERM);
        /* Wait for killed process: try few times waiting for
           processes stuck in kernel mode */
        if (waitpid(remove_child,&child_status,WNOHANG)>0)
          break;
        sleep(1);
        if (waitpid(remove_child,&child_status,WNOHANG)>0)
          break;
       }
     }

    if ((remove_child = fork()) == 0)
     {
      if (uid!=0)
        lpd_client_remove_job(job->queue->lpd_server,
                              job->queue->lpd_queue,
                              user, job->queue_no);
      else
        lpd_client_remove_job(job->queue->lpd_server,
                              job->queue->lpd_queue,
                              "root", job->queue_no);
      exit(0);
     }
   }

  setregid(-1,eff_gid);
  setreuid(-1,eff_uid);

  /* Get rid of temporary file */
  if (strlen(job->temp_file)>0)
    unlink(job->temp_file);
  /* Pop the job out of the doubly-linked list with this macro*/
    remove_job(job);
 }

DSTRING lpr_wrap(service_request *svreq, char *temp_file, pid_t copy_child)
 {
  DSTRING rtdbuf = NULL;
  spooler_queue *current_queue, *move_queue;
  spooler_job *new_job, *move_job;
  spooler_job *current_job;
  struct hostent *host_data;
  char dotted_addr[32],*last_dot,*host_allow="",*host_deny="";
  struct stat file_info;
  int req_id;
  char *sl_pos;
  int status;
  int access;
  int i;

  /*Look up requested queue*/
 
  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    /* Is this a queue reservation release ? */
    if (svreq->entry != NULL && svreq->reserve != PRINTMI_MOVE_REQ)
      if ((sl_pos = strchr(svreq->entry,'/'))!=NULL)
       {
        if ((strstr(svreq->entry,current_queue->name)!=NULL))
         {
          sl_pos++;
          req_id = atoi(sl_pos);
          if (_prints_debug != 0)
          printf("Resolving reservation %s/%d.\n",current_queue->name,
                 req_id);
           
          for (current_job = current_queue->tail_job;
	       current_job != NULL;
               current_job = current_job->prev_job)
	    if (current_job->queue_no == req_id)
	     {
	      if ((current_job->user_uid != svreq->uid) ||
	          (strcasecmp(current_job->host, svreq->host)!=0))
	       {
                rtdbuf=dsprintf(rtdbuf,
	           "lpr: Cannot release reservation %s: not owner (%s@%s/%d)\n",
                      svreq->entry, current_job->user,
		      current_job->host, current_job->user_uid);
                abort_job(temp_file,copy_child);
                return(rtdbuf);
               }
	      if (copy_child>0)
	       {
                /* Allow for file transfer on the reservation RELEASE
	           (2nd) transaction. This might be dropped in a later
		   revision */
  	        current_job->copy_child = copy_child;
	        current_job->status = rescop;
                strncpy(current_job->orig_file,svreq->file,sizeof(current_job->orig_file));
                strncpy(current_job->temp_file,temp_file,sizeof(current_job->temp_file));
                current_job->file_type = LPD_POSTSCRIPT_TYPE; /* NEED TO FIX THIS */
               }
              else              current_job->status = resrdy;
              current_job->is_reservation = FALSE;
              if (svreq->id_only != 0)
	        rtdbuf=dsprintf(rtdbuf,"%s/%d\n",
                       current_queue->name,current_job->queue_no);
              else
	        rtdbuf=dsprintf(rtdbuf,"Reservation ID %s/%d released.\n",
                       current_queue->name,current_job->queue_no);
              return(rtdbuf);
             }       
         }
        
        continue;
       }

    if (strcasecmp(current_queue->name, svreq->printer) == 0)
     {
      /* Found requested queue. Is it locked ? */	 
      if (current_queue->is_locked)
       {
        rtdbuf=dsprintf(rtdbuf,"lpr: queue %s cannot be accessed at this time:\n%s\n",
                current_queue->name,current_queue->status_msg);
        abort_job(temp_file,copy_child);
        return(rtdbuf);
       }
      
      /* Now enforce access control. By user first */
      access = TRUE;
      if (current_queue->user_access_allow != NULL)
        if (strstr(current_queue->user_access_allow, svreq->user)==NULL)
          access = FALSE;
      if (current_queue->user_access_deny != NULL)
        if (strstr(current_queue->user_access_deny,  svreq->user)!=NULL)
          access = FALSE;
      if (!(access))
       {
        rtdbuf=dsprintf(rtdbuf,"lpr: user %s is not allowed to access queue %s.\n",
                svreq->user,current_queue->name);
        abort_job(temp_file,copy_child);
        return(rtdbuf);
       }

      /* Then check whether the host is allowed */
      if ((current_queue->host_access_allow != NULL)||
          (current_queue->host_access_deny  != NULL))
       {
        if (current_queue->host_access_allow != NULL) 
          host_allow = current_queue->host_access_allow;
        if (current_queue->host_access_deny != NULL)
          host_deny  = current_queue->host_access_deny;
        if ((host_data = gethostbyname(svreq->host))==NULL)
         {
          rtdbuf=dsprintf(rtdbuf,"lpr: cannot resolve %s. Access denied.\n",
                svreq->host);
          abort_job(temp_file,copy_child);
          return(rtdbuf);
         }

        if (current_queue->host_access_allow != NULL) 
          access = FALSE;
        if (strstr(host_allow, host_data->h_name)!=NULL)
          access = TRUE;
        if (strstr(host_deny,  host_data->h_name)!=NULL)
          access = FALSE;
        sprintf(dotted_addr,"%u.%u.%u.%u",(unsigned char)host_data->h_addr[0],
                                          (unsigned char)host_data->h_addr[1],
                                          (unsigned char)host_data->h_addr[2],
                                          (unsigned char)host_data->h_addr[3]);
        /* Scan subnets first */
        while ((last_dot = strrchr(dotted_addr,'.'))!=NULL)
         {
          *(last_dot++) = ',';
          *last_dot     = '\000';
        /*
          printf("Checking <%s>: <%s> allowed.\n",dotted_addr,host_allow);
          printf("Checking <%s>: <%s> denied. \n",dotted_addr,host_deny);
         */
          if (strstr(host_allow,dotted_addr)!=NULL) access = TRUE;
          if (strstr(host_deny,dotted_addr)!=NULL)  access = FALSE;
         }
        sprintf(dotted_addr,"%u.%u.%u.%u",(unsigned char)host_data->h_addr[0],
                                          (unsigned char)host_data->h_addr[1],
                                          (unsigned char)host_data->h_addr[2],
                                          (unsigned char)host_data->h_addr[3]);
        if (strstr(host_allow, dotted_addr)!=NULL)
          access = TRUE;
        if (strstr(host_deny,  dotted_addr)!=NULL)
          access = FALSE;
       }
             
      if (!(access))
       {
        rtdbuf=dsprintf(rtdbuf,"lpr: host %s is not allowed to access queue %s.\n",
                svreq->host,current_queue->name);
        abort_job(temp_file,copy_child);
        return(rtdbuf);
       }

      /* If this is a reservation request, does this queue accept those? */
      if ((svreq->reserve == PRINTMI_RESERVE_RES)&&
          ((current_queue->options & SPOOLER_QUEUE_RESV)==0))
       {
        rtdbuf=dsprintf(rtdbuf,"lpr: queue %s does not accept reservations\n",
	        current_queue->name);
        abort_job(temp_file,copy_child);
	return(rtdbuf);
       }
       	      
      /* Handle "job move" requests */
      if (svreq->reserve == PRINTMI_MOVE_REQ)
       {
        if ((sl_pos = strchr(svreq->entry,'/')) == NULL)
         {
          rtdbuf=dsprintf(rtdbuf, "lpr: Cannot move job %s: not found\n",
                  svreq->entry);
          abort_job(temp_file,copy_child);
          return(rtdbuf);
         }
        *sl_pos='\000';
        sl_pos++;
        req_id = atoi(sl_pos);
        if (_prints_debug != 0)
          printf("Now moving %s/%d to queue %s.\n",svreq->entry,req_id,
                  current_queue->name);

        /* Look up job */
        for (move_queue = head_spooler_queue; move_queue != NULL;
             move_queue = move_queue->next_queue)
         {
          if (strcasecmp(move_queue->name, svreq->entry) == 0)
           {
            for (move_job = move_queue->tail_job; move_job != NULL;
                 move_job = move_job->prev_job)
             {
              if (move_job->queue_no == req_id)
               {
                if ((move_job->user_uid != svreq->uid ||
                     strcasecmp(move_job->host, svreq->host)!=0) &&
                     svreq->uid != 0)
                 {
                  rtdbuf=dsprintf(rtdbuf,
                     "lpr: Cannot move job %s/%d: not owner (%s@%s/%d)\n",
                      svreq->entry, req_id, move_job->user,
                      move_job->host, move_job->user_uid);
                  abort_job(temp_file,copy_child);
                  return(rtdbuf);
                 }

                /* BINGO - do the move. */
                unlink_job(move_job); /* This is a macro */
                
                move_job->queue = current_queue;
                move_job->prev_job = current_queue->tail_job;
                move_job->next_job = NULL;

                if (current_queue->tail_job == NULL)
                 {
                  move_job->queue_no = 1;
                  current_queue->max_id = 1;
                 }
                else
                 {
                  move_job->prev_job->next_job = move_job;
                  current_queue->max_id++;
                  move_job->queue_no = current_queue->max_id;
                 }
                current_queue->last_action = time(0);
                current_queue->tail_job = move_job;
                if (svreq->id_only != 0)
                rtdbuf=dsprintf(rtdbuf,"%s/%d\n",
                        current_queue->name,move_job->queue_no);
                else
                rtdbuf=dsprintf(rtdbuf,"New Job ID is %s/%d\n",
                        current_queue->name,move_job->queue_no);
                return(rtdbuf);
               }
             }
           }
         }
        rtdbuf=dsprintf(rtdbuf, "lpr: Cannot move job %s/%d: not found\n",
                  svreq->entry,req_id);
        abort_job(temp_file,copy_child);
        return(rtdbuf);
       }

      /* Ok, this is a regular new request. Enqueue job. */      
      new_job = (spooler_job *)malloc(sizeof(spooler_job));
      new_job->file_type = LPD_POSTSCRIPT_TYPE; /* NEED TO FIX THIS */
      strncpy(new_job->host,svreq->host,sizeof(new_job->host));
      strncpy(new_job->user,svreq->user,sizeof(new_job->user));
      if (_prints_debug!=0)
        printf("User: <%s> --> <%s> Host <%s> --> <%s>\n",
	        svreq->user,new_job->user,
		svreq->host,new_job->host);
      new_job->user_uid = svreq->uid;
      strncpy(new_job->orig_file,svreq->file,sizeof(new_job->orig_file));
      strncpy(new_job->temp_file,temp_file,sizeof(new_job->temp_file));
      new_job->queue = current_queue;
      new_job->prev_job = current_queue->tail_job;
      new_job->next_job = NULL;
      current_queue->last_action = time(0);
      new_job->copy_child = copy_child;
      if (copy_child>0) new_job->status = copying;
      else              new_job->status = ready;
      new_job->spool_child = -1;
      if (svreq->reserve == PRINTMI_RESERVE_RES)
           new_job->is_reservation = TRUE;
      else new_job->is_reservation = FALSE;
      if (new_job->is_reservation) new_job->status = reserve;
      new_job->copies_to_print = svreq->copies;
      new_job->print_retries = 0;
      if (stat(temp_file,&file_info)!=-1)
          new_job->file_size = file_info.st_size;
      else new_job->file_size = 0;
      if (current_queue->tail_job == NULL)
       {
        new_job->queue_no = 1;
	current_queue->max_id = 1;
       }
      else
       {
        new_job->prev_job->next_job = new_job;
	current_queue->max_id++;
        new_job->queue_no = current_queue->max_id;
       }
        
      current_queue->tail_job = new_job;
      if (svreq->id_only != 0)
        rtdbuf=dsprintf(rtdbuf,"%s/%d\n",
                     current_queue->name,new_job->queue_no);
      else
        rtdbuf=dsprintf(rtdbuf,"Job accepted: ID is %s/%d\n",
                     current_queue->name,new_job->queue_no);
      return(rtdbuf);
     }
   }
  rtdbuf=dsprintf(rtdbuf,"Cannot find printer queue %s.\n",svreq->printer);
  abort_job(temp_file,copy_child);
  return(rtdbuf);
 }

int abort_job(char *temp_file,int copy_child)
 {
  aborted_job *new_aborted_job;

  if (copy_child >0)
   {
    new_aborted_job = (aborted_job *)malloc(sizeof(aborted_job));
    new_aborted_job->next_job = head_aborted_job;
    new_aborted_job->copy_child = copy_child;
    strncpy(new_aborted_job->temp_file,temp_file,sizeof(new_aborted_job->temp_file)-1);
    head_aborted_job = new_aborted_job;
   }
  else if (strlen(temp_file)>0)
    unlink(temp_file);

  return(0);
 }

int clean_up_aborted_jobs()
 {
  aborted_job *job,*prev_job,*next_job;
  int n_finished=0,status,child_status;

  prev_job = NULL;

  for (job=head_aborted_job; job!=NULL; job=next_job)
   {
    next_job = job->next_job;
    status = waitpid(job->copy_child,&child_status,WNOHANG);
    if (status>0)
     {
      if (strlen(job->temp_file)>0) unlink(job->temp_file); 
      if (_prints_debug != 0)
        printf("Child %d ended (%d). Cleaning up <%s>\n",
               job->copy_child,child_status,job->temp_file);
      if (prev_job != NULL) prev_job->next_job = next_job;
      else head_aborted_job = next_job;
      n_finished++;
      free(job);
     }
    else prev_job = job;
   }
  return(n_finished);
 }
      
int terminate_aborted_jobs()
 {
  aborted_job *job,*next_job;
  int n_finished=0,status,child_status;

  for (job=head_aborted_job; job!=NULL; job=next_job)
   {
    next_job = job->next_job;
    kill(job->copy_child, SIGTERM);
    status = waitpid(job->copy_child,&child_status,0);
    if (strlen(job->temp_file)>0) unlink(job->temp_file);
    free(job);
   }
  head_aborted_job = NULL;
  return(n_finished);
 }

      
int isempty_wrap()
 {
  spooler_queue *current_queue;
  spooler_job *current_job;
  int total_jobs=0;

  for(current_queue = head_spooler_queue;
      current_queue != NULL; current_queue = current_queue->next_queue)
   {
    current_job = current_queue->tail_job;
     
    if (current_job!=NULL)
     {
      /* First traverse the list backwards */
      while(current_job!=NULL)
       {
        current_job = current_job->prev_job;
        total_jobs++;
       }
     }
   }
  return(total_jobs);
 }


int do_print_wrap()
 {
  spooler_queue *current_queue;
  spooler_job *current_job;
  spooler_job *save_job,*job1,*job2,*temp_job,*printing_job=NULL;
  static time_t start_time=0;
  char *script;
  int child_status;
  static int can_spool = FALSE;
  static time_t can_spool_updated = 0;
  spooler_queue *start_queue;
  struct stat file_info;
  int first_time = TRUE;
  int spooled_one,need_to_sort_queue;
  int retcod;
  int check_fd;
  char check_buffer[16];

  start_time = time(0);
  current_queue = dp_save_current_queue;

  /* First of all, let's take care of any pending aborted job */
  clean_up_aborted_jobs();

  /* Can we start spooling messages (is the secondary server
     empty?) (This information needs to stay cached for a while */
  if ((start_time - can_spool_updated) > (SERVER_INFO_CACHE_TIME * 60))
   {
    if (_prints_primary != 0)
     {
      if (!(is_alive(_prints_other_server))) can_spool = TRUE;
      else if (enquire(_prints_other_server)>0) can_spool = FALSE;
      else can_spool = TRUE;
      if (_prints_debug!=0&&can_spool==FALSE)
      printf("Cannot spool (server %s is alive and busy).\n",
             _prints_other_server);
     }
    else
     {
      if (!(is_alive(_prints_other_server))) can_spool = TRUE;
      else if (isempty_wrap()>0) 
       {
        can_spool = TRUE;
        /* Safety feature: if we are the secondary and we have been idle with
           pending jobs for too long, and the primary is alive, remove all of
           our jobs so that the primary can resume */
        if (((start_time - spooler_last_action)/60) >= STUCK_SECONDARY_WAIT)
         {
          if (_prints_debug!=0)
           {
            printf("Secondary server has been stuck with pending jobs for at least %d minutes. \n",STUCK_SECONDARY_WAIT);
            printf("Cleaning up all jobs. \n");
           }
          reset_all_jobs();
         }
       }
      else can_spool = FALSE;
      if (_prints_debug!=0&&can_spool==FALSE)
      printf("Cannot spool (server %s is alive and there's no jobs left).\n",
             _prints_other_server);
     }
    can_spool_updated = start_time;
   }

  if (can_spool==FALSE) return(0);

  if (current_queue == NULL) current_queue = head_spooler_queue;
  start_queue = current_queue;
    
  for(;time(0)<(start_time+SPOOL_TIMEOUT);
      current_queue = current_queue->next_queue)
   {
    if (current_queue == NULL) current_queue = head_spooler_queue;
    if (current_queue == NULL) break;

    /* Make sure the queue list is scanned only once per call
       (within the timeout) */
    if (first_time) first_time = FALSE;
    else if (current_queue == start_queue) break;  
         
    if (current_queue->is_locked) continue;

    /* Has the queue been idle for too long ? */
    if (current_queue->idle_monitor > 0)
     {
      if (start_time > (current_queue->last_action +
                       current_queue->idle_monitor) )
       { 
        script = getenv("PRINTMI_IDLEMON_SCRIPT");
        if ((script == NULL)||(_prints_debug != 0)) 
	  fprintf(stderr,"Warning: queue %s has been idle for %d minutes.\n",
	          current_queue->name, current_queue->idle_monitor/60);
        else
         {
          if (current_queue->idlemon_child > 0)
           {
            retcod = waitpid(current_queue->idlemon_child,&child_status,WNOHANG);
            if (retcod != 0) current_queue->idlemon_child = -1;
           }
          else if ((current_queue->idlemon_child=fork())==0) 
           /*This is not *required* to succeed! */
	   {
            exit(execlp(script,script,NULL));
           }
         }
       }
     }
   
    if ((current_job = current_queue->tail_job) == NULL) 
     {
      current_queue->last_action = start_time;
      continue;
     }

    spooled_one=FALSE;
    
    /* Find the first job to print */
    while(current_job->prev_job!=NULL)
      current_job = current_job->prev_job;

    /* Check whether the first job in queue is a reservation
       placeholder */
    if (current_job!=NULL)
      if (current_job->is_reservation)
       {
        if (current_job->status != waiting)
	 {
	  current_job->status = waiting;
	  current_queue->last_action = start_time;
	  if (_prints_debug!=0)
	    printf("Found reservation on queue %s. Waiting.\n",
	           current_queue->name);
	  continue;
	 }
	else if (start_time>(current_queue->last_action + MAX_RESERVE_SECONDS))
	 {
	  if (_prints_debug!=0)
	    printf("Reservation time expired. Removing reservation %s/%d.\n",
	           current_job->queue->name, current_job->queue_no);
          save_job = current_job->next_job;
          /* Get rid of temporary file */
          if (strlen(current_job->temp_file)>0) unlink(current_job->temp_file);
	  remove_job(current_job);
	  current_job = save_job;
	 }
	else spooled_one=TRUE; /*We are still waiting*/
       }

    printing_job = NULL;	        
    need_to_sort_queue = FALSE;
    
    while(current_job!=NULL)
     {
      if (current_job->is_reservation) 
       {
        current_job = current_job->next_job;
        continue;
       }
       
      if (current_job->status == printing)
       {
	 if (current_job->spool_child>0)
	  {
           retcod=waitpid(current_job->spool_child,&child_status,WNOHANG);
       	   if (retcod > 0)
	    {
	     if (_prints_debug != 0)
	      {
	       printf("Job %s/%d spool child exit status: %d.\n",
	              current_job->queue->name, current_job->queue_no,
                      WEXITSTATUS(child_status));
	      }
	    }
	   /* Spool child terminated. Retain job
	      in queue until the remote LPD queue is available again */
	   if (retcod>0)
	    {
	     current_job->spool_child = -1;
             if (WEXITSTATUS(child_status)!=0)
              {
               /* Error forking LPD client: 
                  removing and setting status to "error" */
               current_job->status = error;
               save_job = current_job;
               current_job = current_job->next_job;
               /* In case of a transaction error, retry one time then throw the
                  job away: this will prevent Postscript files with fatal
                  errors from being resubmitted and attempted to be
                  printed too many times */
               if (save_job->print_retries < SPOOLER_QUEUE_MAX_RETRIES)
                {
                 requeue_job(save_job); /* This is a macro */
                 save_job->print_retries++;
                }
               else
                {
                 /* Get rid of temporary file */
                 if (strlen(save_job->temp_file)>0)
                   unlink(save_job->temp_file);		
                 remove_job(save_job); /* This is a macro */
                }
               continue;
              }
	     printing_job = current_job;
	    }
	  }
	 else printing_job = current_job;
	 current_job = current_job->next_job;
	 continue;
       }

/*
      if (_prints_debug != 0)        
        printf("Now testing job %s.\n",current_job->orig_file);
 */
 
      /* Did the child copy job complete? */
      if (current_job->copy_child > 0 )
       {
        retcod=waitpid(current_job->copy_child,&child_status,WNOHANG);
        if (retcod<0)
         {
          if (_prints_debug != 0)
            printf("waitpid for PID %d returns %d: %s\n",current_job->copy_child,
                 retcod, strerror(errno));
          current_job = current_job->next_job;
          continue;
         }
	else if (retcod==0)
	 {
          /* Data copy not completed: skip file */
	  current_job = current_job->next_job;
	  continue;
	 }
	
        /* Compute final file size and set a flag to reshuffle
	   the queue if necessary */
	if (stat(current_job->temp_file,&file_info)<0)
	 {
	  /* Cannot stat temp file. Remove bogus job! */
          save_job = current_job;
	  fprintf(stderr,"Cannot stat temp file %s. Removing job %s/%d\n",
	          current_job->temp_file, current_job->queue->name,
		  current_job->queue_no);
	  current_job = current_job->next_job;
	  remove_job(save_job);
	  continue;
	 }
	need_to_sort_queue = TRUE;	  
        current_job->copy_child = -1;
	current_job->file_size = file_info.st_size;
        if (current_job->status == rescop) current_job->status = resrdy;
	else                               current_job->status = ready;
       }


     /* Don't attempt to send jobs via lpr if the queue is marked
         busy or if we already sent one job to this queue during
	 the current cycle */	 
      if ((start_time<current_queue->busy_until) || spooled_one)
       {
        current_job = current_job->next_job;
        continue;
       }
      if (lpd_queue_is_busy(current_queue))                
       {
        spooled_one = TRUE;  /* We don't attempt to print any other job
                                for this queue */
        current_job = current_job->next_job;
	continue;
       }

      /* If we reach here, the queue is available */

      /* If the queue is available, and it was supposed to be printing a
         job, then that should be finished... If the fork process reported
         an error, this is trapped just after the waitpid call above. */
      if (printing_job!=NULL)
       {
   	printing_job->copies_to_print--;
        /* This is a successful job. Let's generate an accounting
           printout */
        if (((printing_job->queue->options)&SPOOLER_QUEUE_ACCT)!=0) 
          printf("ACCT: %s\n",get_accounting_string(printing_job));

	if (printing_job->copies_to_print <=0)
	 {
          /* Get rid of temporary file */
          if (strlen(printing_job->temp_file)>0) 
            unlink(printing_job->temp_file);
	  remove_job(printing_job);  /*This is a macro*/
	 }
	else
	 {
          if (_prints_debug != 0) 
	    printf("%s/%d needs to print %d copies\n",
	            printing_job->queue->name,
		    printing_job->queue_no,
		    printing_job->copies_to_print);
          requeue_job(printing_job); /* This is also a macro */
	 }
        printing_job = NULL;
       }

      /* Check the file for queues that are marked "Postscript Only" */
      if (((current_job->queue->options)&SPOOLER_QUEUE_JPSF)!=0)
       {
        check_fd = open (current_job->temp_file,O_RDONLY);
        if (check_fd < 0)
         {
          /* Cannot open temp file. Remove bogus job! */
          save_job = current_job;
          fprintf(stderr,"Cannot open temp file %s (%s). Removing job %s/%d\n",
                  current_job->temp_file, strerror(errno),
                  current_job->queue->name, current_job->queue_no);
          current_job = current_job->next_job;
          remove_job(save_job);
          continue;
         }
        read(check_fd,check_buffer,sizeof(check_buffer));
        if (strncasecmp(check_buffer,SPOOLER_QUEUE_PS_CHECK,strlen(SPOOLER_QUEUE_PS_CHECK))!=0)
         {
          /* No PS signature in file. Remove job! */
          save_job = current_job;
          fprintf(stderr,"Not postscript (%c%c%c%c%c%c): Removing job %s/%d (%s@%s)\n",
                  check_buffer[0], check_buffer[1], check_buffer[2],
                  check_buffer[3], check_buffer[4], check_buffer[5],
                  current_job->queue->name, current_job->queue_no,
                  current_job->user, current_job->host);
          current_job = current_job->next_job;
          close(check_fd);
          unlink(save_job->temp_file);
          remove_job(save_job);
          continue;
         }
        close(check_fd);
       }

      /* Last minute check: is the file still there? */
      if (stat(current_job->temp_file,&file_info)<0)
       {
        /* Cannot stat temp file. Remove bogus job! */
        save_job = current_job;
        fprintf(stderr,"Temp file %s disappeared! Removing job %s/%d\n",
                current_job->temp_file, current_job->queue->name,
                current_job->queue_no);
        current_job = current_job->next_job;
        remove_job(save_job);
        continue;
       }

      /* Debug printout */
      if (_prints_debug != 0)
       {
        printf("Forking the following LPD job:\n");
        printf("Queue: %s Entry: %d\n", current_job->queue->name,
                                      current_job->queue_no);
        printf("User: %s@%s (%d)\n", current_job->user,
                                   current_job->host,current_job->user_uid);
        printf("File: %s Temp: %s (%d bytes) Typ:%d\n", current_job->orig_file,
                       current_job->temp_file, current_job->file_size,
		       current_job->file_type);
       }
    
      /* Pass queue/job by value to fork child */
      current_job->spool_child = fork_file_spool(*current_queue, *current_job);
      if (current_job->spool_child<0) current_job->status = error;
      else
       {
        current_job->status = printing;
	spooled_one = TRUE;
        current_queue->busy_until = start_time + QUEUE_FREEZE_MIN +
             (current_job->file_size/0x32000)*QUEUE_FREEZE_ADDTL;
	current_queue->last_action = start_time;
        spooler_last_action = start_time;
       }
       
      current_job = current_job->next_job;
     } /* This is the end of the main current_job loop */

    /* Do we need to remove a dangling printing job? */
    if (printing_job!=NULL)
     {
      if (!(lpd_queue_is_busy(current_queue)))
       {
        printing_job->copies_to_print--;
        /* This is a successful job. Let's generate an accounting
           printout */
        if (((printing_job->queue->options)&SPOOLER_QUEUE_ACCT)!=0)
          printf("ACCT: %s\n",get_accounting_string(printing_job));

	if (printing_job->copies_to_print <=0)
	 {
          /* Get rid of temporary file */
          if (strlen(printing_job->temp_file)>0)
            unlink(printing_job->temp_file);
	  remove_job(printing_job);  /*This is a macro*/
	 }
	else
	 {
          if (_prints_debug != 0) 
	    printf("%s/%d needs to print %d copies\n",
	            printing_job->queue->name,
		    printing_job->queue_no,
		    printing_job->copies_to_print);
          requeue_job(printing_job); /* This is also a macro */
	 }
        printing_job = NULL;
       }
     }
    /* Reorder queue by size if new size info is available */
    if (need_to_sort_queue&&((current_queue->options&SPOOLER_QUEUE_SORT)!=0))
     {
      for(current_job = current_queue->tail_job; current_job!=NULL;
          current_job = current_job->prev_job)
        for (save_job = current_job->prev_job; save_job!=NULL;
	     save_job = save_job->prev_job)
	 {
	  job1 = save_job;
	  job2 = save_job->next_job;
          /* Only jobs that are both in the READY state can be swapped */
          /* Jobs that are released after a reservation are in the */
          /* "resrdy" state and thus cannot be moved */
	  if ((job1->status != ready) ||
	      (job2->status != ready)) continue;
	  if (job2->file_size < job1->file_size) 
	   {
	    /*Now swap jobs*/
	    if (_prints_debug != 0)
	     {
	      printf("Before:\n");
	      printf("Job 1: [%08X] next: [%08X] prev [%08X]",
	             job1, job1->next_job, job1->prev_job);
	      if (job1->prev_job != NULL) printf("  prev->next [%08X]",
	                                         job1->prev_job->next_job);
	      printf("\nJob 2: [%08X] next: [%08X] prev [%08X]",
	             job2, job2->next_job, job2->prev_job);
	      if (job2->next_job != NULL) printf("  next->prev [%08X]",
	                                         job1->next_job->prev_job);
              printf("\nCurrent: [%08X] Save:[%08X]\n",current_job,save_job);
	      printf("\n");
	     }		 	    
	    if (job2->next_job!=NULL)
	       job2->next_job->prev_job = job1;
	    if (job1->prev_job!=NULL)
	       job1->prev_job->next_job = job2;
	    job2->prev_job = job1->prev_job;
	    job1->next_job = job2->next_job;
	    job2->next_job = job1;
	    job1->prev_job = job2;
	    
	    if (job2->queue->tail_job == job2)
	           job2->queue->tail_job = job1;
            if (current_job == job2) current_job = job1;
	    save_job = job2;
	    if (_prints_debug != 0)
	     {
	      printf("After:\n");
	      printf("Job 1: [%08X] next: [%08X] prev [%08X]",
	             job1, job1->next_job, job1->prev_job);
	      if (job1->next_job != NULL) printf("  next->prev [%08X]",
	                                         job1->next_job->prev_job);
	      printf("\nJob 2: [%08X] next: [%08X] prev [%08X]",
	             job2, job2->next_job, job2->prev_job);
	      if (job2->prev_job != NULL) printf("  prev->next [%08X]",
	                                         job1->prev_job->next_job);
              printf("\nCurrent: [%08X] Save:[%08X]\n",current_job,save_job);
	      printf("\n");
	     }		 	    
           }
	 }
	 
       if (_prints_debug!=0)
         printf("Re-sorted queue %s\n",current_queue->name);
     }
   }
  dp_save_current_queue = current_queue;
  return(0);
 }

pid_t fork_file_spool(spooler_queue queue_to_spool, spooler_job job_to_spool)
 {
  pid_t child_process;
  int banner;
  int retcod;
  
  if ((queue_to_spool.options&SPOOLER_QUEUE_BANN)!=0) banner=TRUE;
  else                                                banner=FALSE;
  
  if ((child_process = fork())<0)
   {
    perror("Cannot fork file spool process");
    return(-1);
   }

  if (child_process == 0)
   {
    
   /* restore 0 UID to be able to open LPD port */
    if (setreuid(-1,0)!=0)
     fprintf(stderr,"Cannot set UID for LPD port access\n");
    if (setregid(-1,0)!=0)
     fprintf(stderr,"Cannot set GID for LPD port access\n");

    /* This pause takes care of certain HP LPD daemons that will time out
       when the data connection is initiated shortly after the status
       connection */
    if (queue_to_spool.sleep_before_spool > 0)
      sleep(queue_to_spool.sleep_before_spool);

    if (queue_to_spool.telnet_port == 0)
      retcod = lpd_client_spool_file(queue_to_spool.lpd_server,
                          queue_to_spool.lpd_queue,
                          job_to_spool.host,
                          job_to_spool.user,
			  job_to_spool.orig_file,
			  job_to_spool.temp_file,
			  job_to_spool.file_type,
			  job_to_spool.queue_no,
			  banner);
    else  retcod = straight_client_spool_file(queue_to_spool.lpd_server,
                          queue_to_spool.telnet_port, job_to_spool.temp_file,
                          queue_to_spool.sleep_before_close);

    if (retcod<0) exit(255);

    exit(0);
   }

  return(child_process);
  
 }
 
int lpd_queue_is_busy(spooler_queue *current_queue)
 {
  char query_buffer[1024];
  time_t query_time;
  uid_t eff_uid;
  gid_t eff_gid;
  int socket;
  char *tail;
  int i,msg_len;

  if (current_queue->telnet_port != 0)
   {
    if (straight_client_get_status(current_queue->lpd_server,
                                   current_queue->telnet_port)<0)
         return(TRUE);
    else return(FALSE);
   }

  eff_uid = geteuid();
  eff_gid = getegid();

  if (setreuid(-1,0)!=0) fprintf(stderr,"Cannot set UID for LPD port access\n");
  if (setregid(-1,0)!=0) fprintf(stderr,"Cannot set GID for LPD port access\n");
  
  if (lpd_client_get_queue_state(current_queue->lpd_server,
                                 current_queue->lpd_queue,query_buffer,
			         sizeof(query_buffer)) < 0)
   {
    if (errno!=EADDRINUSE)
     {
      query_time = time(0);
      if (_prints_debug != 0)
        printf("Cannot access queue %s on %s on %s",
                      current_queue->lpd_queue,  
	  	      current_queue->lpd_server, ctime(&query_time));
      current_queue->busy_until = query_time + DEAD_QUEUE_BUSY_SECONDS;
     }
    setregid(-1,eff_gid);
    setreuid(-1,eff_uid);
    return(TRUE);
   }
       
  setregid(-1,eff_gid);
  setreuid(-1,eff_uid);

  if (_prints_debug != 0)
    printf("Status from queue %s:\n<%s>\n",
         current_queue->lpd_queue,query_buffer);

  /* Save the first couple of lines of printer status into the queue struct */
  tail = strchr(query_buffer,'\n');
  if (tail != NULL) tail = strchr(tail+1,'\n');
  if (tail != NULL) msg_len = (int)(tail - query_buffer);
  else msg_len = strlen(query_buffer);

  msg_len = msg_len < (sizeof(current_queue->status_msg)-1) ?
            msg_len : (sizeof(current_queue->status_msg)-1);
  strncpy(current_queue->status_msg, query_buffer, msg_len);
  current_queue->status_msg[msg_len] = '\000';

  /* The following is the real weakness of this piece of code. We assume
     that empty lpd servers give a certain response!
     This may need to be changed. */

  if (((current_queue->options)&SPOOLER_QUEUE_NSTA) != 0)
   {
    /* An empty status string is good enough for this printer */
    for (i=0;i<strlen(query_buffer);i++)
     if (query_buffer[i]!=' ' && query_buffer[i]!='\t') break;   
    if (i==strlen(query_buffer)) return(FALSE); /* Empty or blank string */
   }

  if (current_queue->empty_check_string != NULL)
    if (strstr(query_buffer,current_queue->empty_check_string)!=NULL) return(FALSE);
  if (strstr(query_buffer,"No entries")!=NULL) return(FALSE);
  if (strstr(query_buffer,"no entries")!=NULL) return(FALSE);
  if (strstr(query_buffer,"no jobs queued")!=NULL) return(FALSE);
  if (strstr(query_buffer,"Printer OK")!=NULL) return(FALSE);
  if (strstr(query_buffer,"ENABLED")!=NULL) return(FALSE);
  if (strstr(query_buffer," 0 jobs")!=NULL) return(FALSE);
  if (strstr(query_buffer,"On line")!=NULL) return(FALSE);
  if (strstr(query_buffer,"queue empty")!=NULL) return(FALSE);
  if ((strstr(query_buffer,"PrinterWarning")!=NULL) &&
      (strstr(query_buffer,"waiting")==NULL)) return(FALSE);

  /* Now handle DEC Printserver response */
#define PRINTSERVER_TAG "contains:"
  if ((tail=strstr(query_buffer,PRINTSERVER_TAG))!=NULL) 
   {
    tail+=strlen(PRINTSERVER_TAG);
    for (i=0;i<strlen(tail);i++)
     if (*tail != ' ' && *tail != '\r' && *tail != '\n' && *tail != '\t') break;
    if (i == strlen(tail)) return(FALSE);
   }
  return(TRUE);
 }

char *get_accounting_string(spooler_job *job)
 {
  FILE *check_file_fd;
  char line[256];
  int line_index=0,n_pages=0;
  char c,*npage_arg;
  static char *response=NULL;

#define PRINTS_NPAGES_DIRECTIVE "%%Pages:"
#define PRINTS_NPAGE_DIRECTIVE "%%Page:"

  if ((check_file_fd = fopen(job->temp_file, "r")) != NULL)
   {
    while (!feof(check_file_fd))
     {
      c=fgetc(check_file_fd);
      if (c=='\r') continue;
      if (c=='\n'||c==EOF)
       {
        line[line_index]='\000';
        line_index=0;

        if (strncmp(line,PRINTS_NPAGES_DIRECTIVE,
                  strlen(PRINTS_NPAGES_DIRECTIVE))==0)
         {
          if (strstr(line,"atend") != NULL) continue;
          npage_arg = line + strlen(PRINTS_NPAGES_DIRECTIVE);
          sscanf(npage_arg,"%d",&n_pages);
          break;
         }
        if (strncmp(line,PRINTS_NPAGE_DIRECTIVE,
                  strlen(PRINTS_NPAGE_DIRECTIVE))==0)
         {
          npage_arg = line + strlen(PRINTS_NPAGE_DIRECTIVE);
          sscanf(npage_arg,"%d",&n_pages);
         }
        continue;
       }
      if (c==EOF) break;
      if (line_index<(sizeof(line)-1)) line[line_index++] = c;
     }

    if (n_pages>0)
     {
      response = realloc(response,60 + strlen(job->host)+strlen(job->user));
      if (response!=NULL)
        sprintf(response,"%-20s | %s@%s | %4d\n",
                job->queue->name,job->user,job->host,n_pages);
     }

    fclose(check_file_fd);
   }
  else fprintf(stderr,"Cannot open temp file %s: %s\n",job->temp_file,
                        strerror(errno));
  return(response);
 }

