/*
  File :     spooler_queue.h


  Author :   Francesco Prelz ($Author: root $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
   9-Oct-1997 Original release
  10-Sep-1998 Added prototypes for new option testing routines.
  16-Sep-1998 Modified queue structure in order to accommodate
              for a lock flag and status message.
   5-Nov-1998 Added queue alias name support (get_alias_switches,
              dealias_queue_name, SPOOLER_QUEUE_HAS_ALIASES).
   8-Feb-1999 Added extern for head_spooler_queue, so that it can
              be shared by banner code.
   6-Sep-1999 Added "sleep_before_spool" queue parm to accommodate the
              behaviour of newer HP LPD servers.
  23-Jan-2004 Added queue data members to accommodate extra
              arguments to be passed to the 'gs' utility.
              Added extra_gs_arguments prototype.
              Added spooler_queue_free prototype.
  12-Jun-2006 Added queue option bit (SPOOLER_QUEUE_NO_LEVEL1)
              to accommodate printers that won't digest PS Level 1 
              statusdict directives.
   9-Jan-2007 Added empty_check_string.
  27-Nov-2007 Added media_type queue attribute and function.
  11-Feb-2008 Added sleep_before_close option.
  20-Jul-2010 Added gs_output_filter option.

  Description: Common macros and prototypes for the spooler
               queue services (see spooler_queue.c)
               The "job" linked list structure is defined here.
*/

#ifndef SpoolerQueueH
#define SpoolerQueueH

#include "printd.h"

#define RFC1179_MAX_FILE 131
#define RFC1179_MAX_HOST 31
#define RFC1179_MAX_USER 31

#define MAX_QUEUE_NAME_LEN 64
#define MAX_ALIAS_OPTION_LEN 80
#define MAX_TOTAL_QUEUES   100

#define MAX_RESERVE_SECONDS 600
#define DEAD_QUEUE_BUSY_SECONDS 20

#define JAVA_CLIENT_UID 999999

typedef struct aborted_job_s {
     pid_t copy_child;
     char temp_file[FILENAME_MAX];
     struct aborted_job_s *next_job;
} aborted_job;

typedef struct spooler_job_s {
     int  file_type;
     int  queue_no;
     enum {copying, reserve, rescop, resrdy, ready,
           error, waiting, printing} status;
     char host[RFC1179_MAX_HOST];
     char user[RFC1179_MAX_USER];
     uid_t user_uid;
     char orig_file[RFC1179_MAX_FILE];
     char temp_file[FILENAME_MAX];
     pid_t copy_child;
     pid_t spool_child;
     unsigned int file_size;
     unsigned int is_reservation;
     unsigned int copies_to_print;
     unsigned int print_retries;
     struct spooler_queue_s *queue;
     struct spooler_job_s *prev_job;
     struct spooler_job_s *next_job;
 } spooler_job;

/* Macro to destroy a spooler job */
#define remove_job(current_job) \
        unlink_job(current_job); \
        free(current_job);

#define unlink_job(current_job) \
    if (current_job->prev_job != NULL) \
             current_job->prev_job->next_job = current_job->next_job; \
    if (current_job->next_job != NULL) \
             current_job->next_job->prev_job = current_job->prev_job; \
    if (current_job->queue->tail_job == current_job) \
             current_job->queue->tail_job = current_job->prev_job;

/* Macro to put a spooler job at the bottom of the queue */
#define requeue_job(current_job) \
    current_job->status = ready; \
    if (current_job->queue->tail_job != current_job) \
     { \
      if (current_job->next_job != NULL) \
        current_job->next_job->prev_job = current_job->prev_job; \
      if (current_job->prev_job != NULL) \
        current_job->prev_job->next_job = current_job->next_job; \
      current_job->next_job = NULL; \
      current_job->queue->tail_job->next_job = current_job; \
      current_job->prev_job = current_job->queue->tail_job; \
      current_job->queue->tail_job = current_job; \
     } 

typedef struct spooler_queue_s {
     int is_locked;
     char name[MAX_QUEUE_NAME_LEN];
     char lpd_queue[MAX_QUEUE_NAME_LEN];
     char lpd_server[MAXHOSTNAMELEN];
     char status_msg[MAX_STATUS_MESSAGE_LEN];
     char *gs_option;
     char **gs_extra_arguments;
     char *gs_output_filter;
     int  gs_extra_arguments_n;
     struct spooler_queue_s *next_queue;
     struct spooler_job_s   *tail_job;
     char *user_access_allow;
     char *user_access_deny;
     char *host_access_allow;
     char *host_access_deny;
     char *file_check_string;
     int telnet_port;
     char *description;
     int options;
     unsigned int max_id;
     int idlemon_child;
     int sleep_before_spool;
     time_t last_action;
     time_t busy_until;
     time_t idle_monitor;
     char *empty_check_string;
     char *media_type;
     int sleep_before_close;
 } spooler_queue;

typedef struct alias_queue_s {
     char name[MAX_QUEUE_NAME_LEN];
     struct spooler_queue_s *real_queue;
     char switches[MAX_ALIAS_OPTION_LEN];
     struct alias_queue_s *next_queue;
 } alias_queue;

/* Bit defs for the queue "Options" field */
#define SPOOLER_QUEUE_SORT 1
#define SPOOLER_QUEUE_RESV 2
#define SPOOLER_QUEUE_BANN 4
#define SPOOLER_QUEUE_DUPL 8
#define SPOOLER_QUEUE_ACCT 16
#define SPOOLER_QUEUE_NSTA 32
#define SPOOLER_QUEUE_JPSF 64
#define SPOOLER_QUEUE_NODU 128
#define SPOOLER_QUEUE_NOTR 256
#define SPOOLER_QUEUE_LPDS 512
#define SPOOLER_QUEUE_HAS_ALIASES 1024
#define SPOOLER_QUEUE_NO_LEVEL1 2048

#define SPOOLER_QUEUE_PS_CHECK "%!PS-Adobe"

typedef struct control service_request;

/* Prototypes */
int initialize_queues(char *file_name);
pid_t fork_file_spool(spooler_queue queue_to_spool, spooler_job job_to_spool);
int lpd_queue_is_busy(spooler_queue *queue);
int is_duplex(char *queue_name);
int is_noduplex(char *queue_name);
int is_notray(char *queue_name);
int is_nosize(char *queue_name);
int is_nolevel1(char *queue_name);
int wants_banner(char *queue_name);
char *dealias_queue_name(char *name);
char *get_alias_switches(char *name);
char *get_accounting_string(spooler_job *job);
char  *use_gs(char *queue_name);
char **extra_gs_arguments(char *queue_name, int *n_arguments);
char  *gs_output_filter(char *queue_name);
char **parse_arguments(char *start, char *end, int *n_arguments);
char  *check_file(char *queue_name);
char  *media_type(char *queue_name);
int abort_job(char *temp_file,int copy_child);
int clean_up_aborted_jobs();
int terminate_aborted_jobs();
void spooler_queue_free(spooler_queue *q);
void spooler_queue_reload();
void spooler_exit_cleanup();
void do_dequeue_and_remove_job(spooler_job *job, int uid, char *user);
void reset_all_jobs();

/* Maximum street time used by do_print_wrap() (seconds) */
#define SPOOL_TIMEOUT 3
/* Minumum time between updates of secondary server activity (minutes) */
#define SERVER_INFO_CACHE_TIME 3
/* Minimum queue freeze time after each job is spooled (seconds) */
#define QUEUE_FREEZE_MIN 10
/* Additional queue freeze time (per Mbyte of spooled data) (seconds) */
#define QUEUE_FREEZE_ADDTL 2
/* Time (minutes) that a secondary server will wait before resetting its jobs
   when there are jobs that cannot be printed and the primary server is 
   alive again. */
#define STUCK_SECONDARY_WAIT (SERVER_INFO_CACHE_TIME*3)
/* Maximum number of times a print job will be resubmitted when errors
   occur during the spooling tothe final queue/service */
#define SPOOLER_QUEUE_MAX_RETRIES 1
 
extern spooler_queue *head_spooler_queue;

#endif /*SpoolerQueueH*/
