/*
  File :     straight_client.c


  Author :   Francesco Prelz ($Author: printmi $)
  e-mail :   "francesco.prelz@mi.infn.it"

  Revision history :
   8-Jun-1998 Original release
  11-Feb-2008 Added sleep_before_close option.
   3-Aug-2009 Adapted to support IPv6 via getaddrinfo.

  Description: File transfer routines into a printer TCP port.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>

#include "straight_client.h"

int straight_client_get_status(char *print_server, int port) 
 {
  int socket;

  socket = get_straight_connection(print_server, port);

  if (socket < 0) return(socket);

  close(socket);
  return(0); 
 }

int straight_client_spool_file(char *print_server, int port, char *name,
                               int sleep_before_close) 
 {
  int file_to_spool;
  struct stat file_stat;
  int socket;
  int read_length,write_length,total_length=0;
  unsigned char buffer[8192];
  int retcod;
  char zero='\000';


  if ((file_to_spool = open(name,O_RDONLY))==-1)
   {
    fprintf(stderr,"Cannot open file %s:%s\n",name,strerror(errno));
    return(-1);
   }
   
  if (fstat(file_to_spool,&file_stat)==-1)
   {
    fprintf(stderr,"Cannot get info about file %s:%s\n",name,strerror(errno));
    return(-1);
   }
   
  if ((socket = get_straight_connection(print_server,port))<0)
   {
    fprintf(stderr,"Cannot open port %d on server %s:%s\n",port,print_server,
                                                           strerror(errno));
    close(file_to_spool);
    return(-1);
   }

  while((read_length=read(file_to_spool,buffer,sizeof(buffer)))>0)
   {
    write_length = write(socket,buffer,read_length);
    if (write_length < 0)
     {
      fprintf(stderr,"Writing to printer socket (port %d) on %s: %s\n",port,
              print_server,
              strerror(errno));
      close(socket);
      return(-1);
     }
    total_length+=write_length;
   }

  close(file_to_spool);

  if (sleep_before_close > 0)
   {
    sleep(sleep_before_close);
   }

  close(socket);
   
  if (total_length!=file_stat.st_size) fprintf(stderr,"File %s: size %d sent %d\n",
                                               name,file_stat.st_size,total_length);
 }

int get_straight_connection(char *print_server, int port)
 {
  int fd_socket;
  int status,original_status;
  static char *last_resolution=NULL;
  static struct addrinfo *ai_ans=NULL, *last_ans=NULL;
  struct addrinfo ai_req, *cur_ans; 
  char port_string[16];
  int retcod;
  fd_set writefs;
  struct timeval timeout;

  /* Get destination address */

  if (last_resolution == NULL || (strcmp(last_resolution, print_server) != 0))
   { 
    if (last_resolution != NULL)
     {
      free(last_resolution);
      freeaddrinfo(ai_ans);
      last_resolution = NULL;
      ai_ans = NULL;
     }
    ai_req.ai_flags = 0;
    ai_req.ai_family = PF_UNSPEC;
    ai_req.ai_socktype = SOCK_STREAM;
    ai_req.ai_protocol = 0; /* Any protocol is OK */
    snprintf(port_string,sizeof(port_string),"%5d",port);
    if ((retcod = getaddrinfo(print_server, port_string, &ai_req, &ai_ans)) != 0)
     {
      fprintf(stderr, "getaddrinfo call for %s failed. return code = %d\n",
              print_server, retcod);
      return(-1);
     }  
   }

  if (last_resolution != NULL)
   {
    cur_ans = last_ans;
   }
  else cur_ans = ai_ans;

  /* Search for a working transport */
  for(;cur_ans != NULL; cur_ans=cur_ans->ai_next)
   {
    if ((fd_socket=socket(cur_ans->ai_family, cur_ans->ai_socktype,
                          cur_ans->ai_protocol))==-1) continue;

    /* Make socket asynchronous in order to enforce connect timeout */

    if ((original_status=fcntl(fd_socket,F_GETFL,NULL)) == -1)
     {
      perror("Cannot fcntl (GETFL) origin socket");
      close(fd_socket);
      continue;
     }
    status = original_status|FNDELAY;
    if ((fcntl(fd_socket,F_SETFL,status)) == -1)
     {
      perror("Cannot fcntl (SETFL) origin socket");
      close(fd_socket);
      continue;
     }

    if (connect(fd_socket,cur_ans->ai_addr,cur_ans->ai_addrlen)==-1)
     {
      if (errno!=EINPROGRESS)
       {
        close(fd_socket);
        continue;
       }
     }

    FD_ZERO(&writefs);
    FD_SET(fd_socket,&writefs);
    timeout.tv_sec=STRAIGHT_CONNECT_TIMEOUT;
    timeout.tv_usec=0;
    if (select(FD_SETSIZE,(fd_set *) 0,&writefs,(fd_set *) 0,&timeout) <= 0)
     {
      close(fd_socket);
      continue;
     }  

    /* restore socket (synchronous) status */
    if ((fcntl(fd_socket,F_SETFL,original_status)) == -1)
     {
      perror("Cannot fcntl (SETFL) origin socket");
      close(fd_socket);
      continue;
     }

    /* Everything's OK. Let's save the contact info. */
    last_resolution = strdup(print_server);
    if (last_resolution != NULL)
     {
      last_ans = cur_ans;
     }
    return(fd_socket);
  }

  freeaddrinfo(ai_ans);
  ai_ans = NULL;
  last_resolution = NULL;
  return(-1);
 }

